#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <unistd.h>
#include <libgnomevfs/gnome-vfs.h>
#include <vte/vte.h>

#include <gtksourceview/gtksourceview.h>
#include <gtksourceview/gtksourcelanguage.h>
#include <gtksourceview/gtksourcelanguagesmanager.h>
#include <gtksourceview/gtksourcetag.h>

#include "debug.h"
#include "editor.h"
#include "main.h"
#include "if_world.h"
#include "net.h"
#include "g_options.h"
#include "mcp/mcp_simpleedit.h"
#include "term.h"
#include "support.h"
#include "world.h"

void on_tlbSave_clicked(GtkButton * button, editor_info * e);
void on_tlbSaveClose_clicked(GtkButton * button, editor_info * e);
void on_tlbClose_clicked(GtkButton * button, editor_info * e);

void editor_save_lines(editor_info * e);

static GList *editors = NULL;
static unsigned int uniqueFileCounter = 0;

static GtkSourceLanguage *lang = NULL;
static GtkSourceLanguagesManager *lm = NULL;

void
on_editor_file_changed(GnomeVFSMonitorHandle * handle,
                       const gchar * monitor_uri, const gchar * info_uri,
                       GnomeVFSMonitorEventType event_type, editor_info * e) {
  switch (event_type) {
  case GNOME_VFS_MONITOR_EVENT_CHANGED:
  case GNOME_VFS_MONITOR_EVENT_CREATED:
    // Aah, enlightenment. Upload this one!
    debug_msg(1, "on_editor_file_changed: change event detected!");
    editor_save_lines(e);
    break;
  default:
    break;
  }
}

void
on_vte_editor_exited(VteTerminal * vteterminal, editor_info * e) {
  editor_close(e);
}

void
on_editor_exited(GPid pid, gint status, editor_info * e) {
  editor_close(e);
}

void
generate_filename(editor_info * e) {
  gchar *filename;
  int i, j;

  if (e->edtype == E_INTERN || !e->name) {
    return;
  }

  if (e->filename) {
    g_free(e->filename);
  }

  filename = g_strdup(e->name);
  j = 0;

  // Remove any characters that are not digits or chars
  // Replace spaces with underscores
  for (i = 0; i < strlen(filename); i++) {
    if (filename[i] == ' ') {
      filename[i] = '_';
    }

    if (!
        (filename[i] != '.' && filename[i] != ':'
         && (filename[i] < '0' || (filename[i] > '9' && filename[i] < 'A')
             || (filename[i] > 'Z' && filename[i] < 'a')
             || filename[i] > 'z'))) {
      filename[j] = filename[i];
      j++;
    }
  }

  filename[j] = '\0';

  if (e->isCode) {
    e->filename =
      g_strdup_printf("%s/tmp/%s.%d.%s", e->owner->path, filename,
                      uniqueFileCounter, "moo");
  } else {
    e->filename =
      g_strdup_printf("%s/tmp/%s.%d.%s", e->owner->path, filename,
                      uniqueFileCounter, "txt");
  }

  uniqueFileCounter++;

  g_free(filename);
}

void
editor_destroy(editor_info * e) {
  GList *line;

  // Free the editor info
  g_free(e->name);
  
  if (e->upload) {
    g_free(e->upload);
  }
  
  if (e->type) {
    g_free(e->type);
  }

  for (line = e->lines; line; line = line->next) {
    g_free((gchar *) (line->data));
  }

  g_list_free(e->lines);

  if (e->edtype == E_EXTERN && e->filename) {
    if (g_file_test(e->filename, G_FILE_TEST_EXISTS)) {
      remove(e->filename);
    }
    g_free(e->filename);
  }

  if (g_list_find(editors, e)) {
    editors = g_list_remove(editors, e);
  }

  g_free(e);
}

void
editor_invoke_oob(world * wld, gchar * line) {
  // parse the line it is formatted: name: blaaaaa upload: jaaaah
  gchar *nameStart, *uploadStart;

  nameStart = strstr(line, "name: ");

  if (!nameStart) {
    return;
  }

  uploadStart = strstr(line, " upload: ");

  if (!uploadStart) {
    return;
  }

  wld->editing = g_new(editing_info, 1);
  wld->editing->name = g_strndup(nameStart + 6, (uploadStart - nameStart) - 6);
  wld->editing->upload = g_strndup(uploadStart + 9, (uploadStart - line) + 9);
  wld->editing->lines = NULL;
}

void
editor_oob_add(world * wld, gchar * line) {
  wld->editing->lines = g_list_append(wld->editing->lines, g_strdup(line));
}

GtkWidget *
editor_create_sourceview(editor_info * e) {
  GtkTextBuffer *buffer;
  GtkWidget *view;
  GtkTextIter iter;
  gchar *line;
  GList *lines;
  PangoFontDescription *f;

  GtkSourceTagStyle *st;

  view = gtk_source_view_new();
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

  if (lang) {
    gtk_source_buffer_set_language(GTK_SOURCE_BUFFER(buffer), lang);
  }

  gtk_text_buffer_get_end_iter(buffer, &iter);

  for (lines = e->lines; lines; lines = lines->next) {
    line = (gchar *) (lines->data);
    gtk_text_buffer_insert(buffer, &iter, line, strlen(line));

    if (lines->next) {
      gtk_text_buffer_insert(buffer, &iter, "\r\n", 2);
    }
  }

  st = gtk_source_tag_style_new();
  st->mask = GTK_SOURCE_TAG_STYLE_USE_BACKGROUND;
  gdk_color_parse("#0000FF", &(st->background));
  gtk_source_buffer_set_bracket_match_style(GTK_SOURCE_BUFFER(buffer), st);

  f =
    pango_font_description_from_string(options_get_str
                                       (main_get_options(),
                                        "font-family"));

  if (f) {
    gtk_widget_modify_font(view, f);
    pango_font_description_free(f);
  }

  gtk_source_view_set_insert_spaces_instead_of_tabs(GTK_SOURCE_VIEW(view),
                                                    TRUE);
  gtk_source_view_set_auto_indent(GTK_SOURCE_VIEW(view), TRUE);
  gtk_source_view_set_show_line_numbers(GTK_SOURCE_VIEW(view), TRUE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), GTK_WRAP_WORD_CHAR);
  gtk_text_view_set_left_margin(GTK_TEXT_VIEW(view), 6);
  gtk_text_view_set_right_margin(GTK_TEXT_VIEW(view), 6);

  if (e->isCode) {
    gtk_source_buffer_set_highlight(GTK_SOURCE_BUFFER(buffer), TRUE);
  } else {
    gtk_source_buffer_set_highlight(GTK_SOURCE_BUFFER(buffer), FALSE);
  }

  gtk_widget_show(view);

  return view;
}

gboolean
editor_write_file(gchar * filename, GList * contents) {
  FILE *f;

  f = fopen(filename, "w");

  if (f) {
    debug_msg(1, "editor_write_file: writing file %s!", filename);
    for (; contents; contents = contents->next) {
      fputs((char *) (contents->data), f);

      if (contents->next) {
        fputs("\n", f);
      }
    }

    fclose(f);
    return TRUE;
  } else {
    debug_msg(1, "editor_write_file: couldn't open file %s for writing!",
              filename);
    return FALSE;
  }
}

gboolean
create_external_editor(editor_info * e, char **argv) {
  GError *err = NULL;

  g_spawn_async(NULL, argv, NULL,
                G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL,
                &(e->pid), &err);

  if (err) {
    debug_msg(1, "create_external_editor: couldn't spawn editor: %s",
              err->message);
    g_error_free(err);
    return FALSE;
  } else {
    return TRUE;
  }
}

void
editor_create(editor_info * e) {
  GladeXML *e_xml = NULL;       // Used to create the editor widget
  GtkWidget *child = NULL;
  GtkWidget *lbl = NULL;
  GtkWidget *view = NULL;
  GtkWidget *srl = NULL;
  term_info *termi = NULL;
  label_info info;
  gchar *name = NULL, *line = NULL;
  GList *elem;
  gchar *icon = NULL, *command = NULL;
  gint i;
  GError *err = NULL;
  const gchar *loc;

  // Find out what type of editor should be created
  if (strcmp(options_get_str(main_get_options(), "editor_alternative"), "0") !=
      0) {
    if (options_get_int(main_get_options(), "editor_embed")) {
      e->edtype = E_VTE;
    } else {
      e->edtype = E_EXTERN;
    }
  } else {
    e->edtype = E_INTERN;
  }

  // Convert the lines to UTF-8 if its our own internal editor thing,
  // otherwise its best to convert it to the current locale!
  g_get_charset(&loc);

  for (elem = e->lines; elem; elem = elem->next) {
    // Convert to UTF-8 if its our intern editor, or if locale is for some reason NULL
    if (e->edtype == E_INTERN || !loc) {
      line =
        g_convert((gchar *) (elem->data), strlen((gchar *) (elem->data)),
                  "UTF-8", options_get_str(e->owner->settings,
                                           "charset"), NULL, NULL, &err);
    } else {
      line =
        g_convert((gchar *) (elem->data), strlen((gchar *) (elem->data)), loc,
                  options_get_str(e->owner->settings, "charset"), NULL,
                  NULL, &err);
    }

    if (line) {
      g_free((gchar *) (elem->data));
      elem->data = line;
    } else {
      debug_msg(1, "editor_create: charset conversion failed: %s",
                err->message);
      g_free(err);
    }
  }

  if (e->edtype == E_EXTERN || e->edtype == E_VTE) {
    // Create the file to be edited
    e->filename = NULL;
    generate_filename(e);

    if (!editor_write_file(e->filename, e->lines)) {
      // Editor cannot proceed!
      editor_destroy(e);
      return;
    }
    // Add a monitor to this file
    gnome_vfs_monitor_add(&(e->filemon), e->filename, GNOME_VFS_MONITOR_FILE,
                          (GnomeVFSMonitorCallback) on_editor_file_changed, e);
    command =
      g_strconcat(options_get_str(main_get_options(), "editor_alternative"),
                  " ", e->filename, NULL);
  }

  if (e->edtype == E_INTERN) {
    e_xml = glade_xml_new(G_XML, "vboxEditor", "wndContainer");
    child = glade_xml_get_widget(e_xml, "vboxEditor");

    srl = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(srl);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(srl),
                                   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    view = editor_create_sourceview(e);
    gtk_container_add(GTK_CONTAINER(srl), view);
    gtk_container_add(GTK_CONTAINER(child), srl);

    e->text = view;
  } else if (e->edtype == E_VTE) {
    termi = term_create_widget(command);
    child = termi->widgets.hbox;
    g_signal_connect((gpointer) termi->widgets.term, "child-exited",
                     G_CALLBACK(on_vte_editor_exited), e);
    g_free(termi);
    g_free(command);
  }

  if (e->edtype == E_INTERN || e->edtype == E_VTE) {
    if (e->isCode) {
      icon = g_strdup("editor_verb.xpm");
    } else {
      icon = g_strdup("editor_text.xpm");
    }

    name =
      g_convert(e->name, strlen(e->name), "UTF-8",
                options_get_str(e->owner->settings, "charset"), NULL,
                NULL, NULL);

    if (name) {
      lbl = create_tab_label(icon, name, TRUE, &info);
      g_free(name);
    } else {
      lbl = create_tab_label(icon, e->name, TRUE, &info);
    }

    g_free(icon);


    i =
      gtk_notebook_append_page(GTK_NOTEBOOK(e->owner->widgets.noteContainer),
                               child, lbl);

    world_scroll_end_prepare(e->owner);

    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(e->owner->widgets.noteContainer),
                               TRUE);
    GTK_WIDGET_UNSET_FLAGS(e->owner->widgets.noteContainer, GTK_CAN_FOCUS);
    gtk_notebook_set_current_page(GTK_NOTEBOOK
                                  (e->owner->widgets.noteContainer), i);

    // Do this so the tabs can become visible, this infects the textview which might be scrolled
    doEvents();
    world_scroll_end(e->owner);

    if (e->edtype == E_INTERN) {
      glade_xml_signal_connect_data(e_xml, "on_tlbSave_clicked",
                                    G_CALLBACK(on_tlbSave_clicked), e);
      glade_xml_signal_connect_data(e_xml, "on_tlbSaveClose_clicked",
                                    G_CALLBACK(on_tlbSaveClose_clicked), e);
      glade_xml_signal_connect_data(e_xml, "on_tlbClose_clicked",
                                    G_CALLBACK(on_tlbClose_clicked), e);
    }

    g_signal_connect((gpointer) info.tabExitButton, "clicked",
                     G_CALLBACK(on_tlbClose_clicked), e);

    e->child = child;
  }

  if (e->edtype == E_EXTERN) {
    char *term = g_strdup(getenv("TERM"));
    char *spawn_command[4] = { NULL, NULL, NULL, NULL };
    char **av;
    int an;

    if (!term) {
      term = g_strdup("xterm");
    }

    if (options_get_int(main_get_options(), "editor_needs_terminal")) {
      spawn_command[0] = term;
      spawn_command[1] = "-e";
      spawn_command[2] = command;

      create_external_editor(e, spawn_command);
    } else {
      g_shell_parse_argv(command, &an, &av, NULL);
      create_external_editor(e, av);
    }

    g_child_watch_add(e->pid, (GChildWatchFunc) on_editor_exited, e);

    g_free(term);
    g_free(command);
  }

  editors = g_list_append(editors, e);
}

void
editor_oob_end(world * wld) {
  editor_info *e = g_new(editor_info, 1);

  e->name = g_strdup(wld->editing->name);
  e->upload = g_strdup(wld->editing->upload);
  e->type = NULL;
  e->etype = E_EXTERNAL;
  e->filename = NULL;
  e->text = NULL;
  e->isCode = (strncmp(e->upload, "@program", 7) == 0);
  e->owner = wld;
  e->lines = g_list_copy(wld->editing->lines);

  g_free(wld->editing->name);
  g_free(wld->editing->upload);
  g_list_free(wld->editing->lines);
  g_free(wld->editing);
  wld->editing = NULL;

  editor_create(e);
}

void
editor_invoke_mcp(world * wld, gchar * name, gchar * reference, gchar * type,
                  GList * lines) {
  editor_info *e = g_new(editor_info, 1);

  e->name = g_strdup(name);
  e->upload = g_strdup(reference);
  e->type = g_strdup(type);
  e->etype = E_MCP;
  e->filename = NULL;
  e->text = NULL;
  e->isCode = (strcmp(type, "moo-code") == 0);
  e->owner = wld;
  e->lines = g_list_copy(lines);

  editor_create(e);
}

GList *
editor_break_line(gchar * line) {
  GList *result = NULL;
  char *pos;

  while ((pos = strstr(line, "\n"))) {
    result = g_list_append(result, g_strndup(line, pos - line));

    line = pos + 1;
  }

  result = g_list_append(result, g_strdup(line));

  return result;
}

GString *
editor_read_file(editor_info * e) {
  FILE *f;
  gchar line[1024];
  GString *str = NULL;

  if (!e->filename) {
    return NULL;
  }

  f = fopen(e->filename, "r");

  if (f) {
    str = g_string_new("");

    while (fgets((char *) &line, 1024 - 1, f) != NULL) {
      str = g_string_append(str, line);
    }

    fclose(f);

    return str;
  } else {
    debug_msg(1, "editor_read_file: file (%s) could not be read: %s",
              e->filename, strerror(errno));
    return NULL;
  }
}

void
editor_save_lines(editor_info * e) {
  // Save the lines
  GtkTextBuffer *buffer;
  GtkTextIter startIter, endIter;
  gchar *lines, *glines;
  GString *line;
  GList *lns = NULL, *elem;
  const gchar *loc;

  if (e->owner->net.state == NET_CONNECTED) {
    if (e->edtype == E_INTERN) {
      buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(e->text));
      gtk_text_buffer_get_end_iter(buffer, &endIter);
      gtk_text_buffer_get_start_iter(buffer, &startIter);
      lines = gtk_text_buffer_get_text(buffer, &startIter, &endIter, FALSE);
    } else if (e->edtype == E_EXTERN || e->edtype == E_VTE) {
      line = editor_read_file(e);

      if (!line) {
        return;
      } else {
        g_get_charset(&loc);
        lines =
          g_convert(line->str, line->len, "UTF-8", loc, NULL, NULL, NULL);

        if (!lines) {
          lines = line->str;
          g_string_free(line, FALSE);
        } else {
          g_string_free(line, TRUE);
        }
      }
    } else {
      return;
    }

    glines =
      g_convert(lines, strlen(lines),
                options_get_str(e->owner->settings, "charset"), "UTF-8",
                NULL, NULL, NULL);

    if (e->etype == E_EXTERNAL) {
      net_send_line(e->owner, e->upload);
    }

    if (glines) {
      if (e->etype == E_EXTERNAL) {
        net_send_line(e->owner, glines);
        g_free(glines);
      } else {
        lns = editor_break_line(glines);
      }
    } else {
      if (e->etype == E_EXTERNAL) {
        net_send_line(e->owner, lines);
      } else {
        lns = editor_break_line(lines);
      }
    }

    if (e->etype == E_EXTERNAL) {
      net_send_line(e->owner, ".");
    } else {
      mcp_simpleedit_save(e->owner, e->upload, e->type, lns);

      for (elem = lns; elem; elem = elem->next) {
        g_free((gchar *) (elem->data));
      }

      g_list_free(lns);
    }

    gtk_notebook_set_current_page(GTK_NOTEBOOK
                                  (e->owner->widgets.noteContainer), 0);
  }
}

void
editor_close(editor_info * e) {
  debug_msg(1, "editor_close: closing editor %s", e->name);

  if ((e->edtype == E_INTERN || e->edtype == E_VTE) && e->owner->loaded &&
      GTK_IS_NOTEBOOK(e->owner->widgets.noteContainer)) {
    // Remove the tab
    gtk_notebook_remove_page(GTK_NOTEBOOK(e->owner->widgets.noteContainer),
                              gtk_notebook_page_num(GTK_NOTEBOOK
                                                    (e->owner->widgets.
                                                    noteContainer), e->child));

    if (gtk_notebook_get_n_pages(GTK_NOTEBOOK(e->owner->widgets.noteContainer))
        == 1) {
      gtk_notebook_set_show_tabs(GTK_NOTEBOOK(e->owner->widgets.noteContainer),
                                  FALSE);
    }
  }

  if (e->edtype == E_EXTERN || e->edtype == E_VTE) {
    // Both use external files wich are monitored, so remove the monitoring
    gnome_vfs_monitor_cancel(e->filemon);
  }

  editor_destroy(e);
}

void
on_tlbSave_clicked(GtkButton * button, editor_info * e) {
  editor_save_lines(e);
}

void
on_tlbClose_clicked(GtkButton * button, editor_info * e) {
  editor_close(e);
}

void
on_tlbSaveClose_clicked(GtkButton * button, editor_info * e) {
  editor_save_lines(e);
  editor_close(e);
}

void
editors_close(world * wld) {
  GList *l = g_list_copy(editors);
  editor_info *e;

  for (l = g_list_copy(editors); l; l = l->next) {
    e = (editor_info *) (l->data);
    if (e->owner == wld) {
      editor_close(e);
    }
  }

  g_list_free(l);
}

void 
setup_tags() {
  options *opt = main_get_options();
  GSList *tags, *tag;
  gchar *name, *oname, *value, *id, **style;
  GtkSourceTag *st;
  GtkSourceTagStyle *sty;
  
  tags = gtk_source_language_get_tags(lang);
  
  for (tag = tags; tag; tag = tag->next) {
    st = GTK_SOURCE_TAG(tag->data);
    sty = gtk_source_tag_get_style(st);
        
    g_object_get(G_OBJECT(st), "name", &name, "id", &id, NULL);
    oname = g_strconcat("editor_", name, NULL);
    
    if ((value = options_get_str(opt, oname))) {
      style = g_strsplit(value, ",", 6);
      
      if (garray_length(style) == 6) {
        /* Foreground color */
        if (strcmp(style[0], "0") == 0) {
          sty->mask = sty->mask & ~GTK_SOURCE_TAG_STYLE_USE_FOREGROUND;
        } else {
          sty->mask = sty->mask | GTK_SOURCE_TAG_STYLE_USE_FOREGROUND;
          gdk_color_parse(style[0], &(sty->foreground));
        }
        
        /* Background color */
        if (strcmp(style[1], "0") == 0) {
          sty->mask = sty->mask & ~GTK_SOURCE_TAG_STYLE_USE_BACKGROUND;
        } else {
          sty->mask = sty->mask | GTK_SOURCE_TAG_STYLE_USE_BACKGROUND;
          gdk_color_parse(style[1], &(sty->background));
        }
        
        sty->bold = atoi(style[2]);
        sty->italic = atoi(style[3]);
        sty->underline = atoi(style[4]);
        sty->strikethrough = atoi(style[5]);
        
        gtk_source_language_set_tag_style(lang, id, sty);
      }
      
      g_strfreev(style);
    }
    
    g_free(oname);
    g_free(name);
    g_free(id);
  }
  
  g_slist_free(tags);
}

void
editor_init() {
  lm = gtk_source_languages_manager_new();
    
  lang =
  gtk_source_languages_manager_get_language_from_mime_type(lm, "text/x-moo");
  
  setup_tags();
}

void
editor_fini() {
  GList *item;
  GList *l = g_list_copy(editors);
  
  for (item = l; item; item = item->next) {
    editor_close((editor_info *)(item->data));
  }
  
  g_list_free(l);
  g_list_free(editors);
  
  if (lm) {
    g_object_unref(G_OBJECT(lm));
  }
}

GtkSourceLanguage *
editor_get_language() {
  return lang;
}
