#include <glib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "g_options.h"
#include "my-string.h"
#include "debug.h"

extern int errno;

// Adds an option to opt, even if key already exists
void
options_add(options * opt, char *key, char *value) {
  char *trimmedKey = mystring_trim(key);

  // Increasing length of ->values with 1
  opt->values =
    (option **) realloc(opt->values,
                        sizeof(option *) * (opt->optionCount + 1));
  // Allocate memory space for the new option
  opt->values[opt->optionCount] = (option *) malloc(sizeof(option));

  // Set key
  opt->values[opt->optionCount]->key = g_strdup(trimmedKey);
  // Set value
  opt->values[opt->optionCount]->value = g_strdup(value);

  opt->optionCount++;
  g_free(trimmedKey);
}

// Sets the value of the given 'key' to 'value'. If key doesn't exists the 'key/value' pair is added
void
options_set_str(options * opt, char *key, char *value) {
  int i;
  char *trimmedKey = mystring_trim(key);

  for (i = 0; i < opt->optionCount; i++) {
    if (g_strcasecmp(opt->values[i]->key, trimmedKey) == 0) {
      // Freeing the old value
      g_free(opt->values[i]->value);
      opt->values[i]->value = g_strdup(value);
      g_free(trimmedKey);
      return;
    }
  }

  // Option wasn't there so we should add it
  options_add(opt, trimmedKey, value);
  g_free(trimmedKey);
}

option *
options_find(options * opt, char *key) {
  int i;

  for (i = 0; i < opt->optionCount; i++) {
    if (g_strcasecmp(opt->values[i]->key, key) == 0) {
      return opt->values[i];
    }
  }

  return NULL;
}

int options_exists(options * opt, char *key) {
  return options_find(opt, key) != NULL;
}

// Returns the value for 'key' if 'key' exists, else NULL
// the returned value is the same as the internal value
// do NOT modify or free it
char *
options_get_str(options * opt, char *key) {
  option *op = options_find(opt, key);

  if (op) {
    return op->value;
  } else {
    return NULL;
  }
}

void
options_set_int(options * opt, char *key, int value) {
  char val[15];

  g_snprintf((char *) &val, 15, "%d", value);
  options_set_str(opt, key, (char *) &val);
}

int
options_get_int(options * opt, char *key) {
  option *op = options_find(opt, key);
  int ret;

  if (op && mystring_to_int(op->value, &ret)) {
    return ret;
  } else {
    return 0;
  }
}

void
options_save(options * opt) {
  FILE *f;
  int i;

  if (opt->filename == NULL) {
    return;
  }

  f = fopen(opt->filename, "w");

  debug_msg(1, "options_save: saving options (%s)!", opt->filename);

  if (f) {
    for (i = 0; i < opt->optionCount; i++) {
      fprintf(f, "%s=%s\n", opt->values[i]->key, opt->values[i]->value);
    }
    fclose(f);
    // Make sure the file is 0660
    chmod(opt->filename, 0660);

  } else {
    debug_msg(1, "options_save: couldn't open option file for saving: %s",
              strerror(errno));
  }
}

options *
options_create() {
  options *result = (options *) malloc(sizeof(options));

  result->optionCount = 0;
  result->values = NULL;
  result->filename = NULL;

  return result;
}

void
options_destroy(options * opt) {
  int i;

  // Free filename
  g_free(opt->filename);

  // Free values
  for (i = 0; i < opt->optionCount; i++) {
    g_free(opt->values[i]->key);
    g_free(opt->values[i]->value);
    g_free(opt->values[i]);
  }

  g_free(opt->values);
  g_free(opt);
}

void
options_load(options * opt, char *filename) {
  FILE *f;
  gchar **keyvalue, line[1024];
  int i;

  debug_msg(1, "options_load: loading options (%s)!", filename);

  if ((f = fopen(filename, "r")) != NULL) {
    i = 0;
    while (fgets((char *) &line, 1024 - 1, f) != NULL) {
      line[strlen((char *) &line) - 1] = '\0';
      i++;
      if (strlen(line) != 0) {  // Empty lines, we don't need to process those
        keyvalue = g_strsplit(line, "=", 2);
        // This will return at least 1 element, at most 2, we need 2
        if (strncmp(keyvalue[0], "#", 1) != 0) {  // Commented lines, well ignore them too
          if (keyvalue[1] != NULL) {
            options_set_str(opt, keyvalue[0], keyvalue[1]);
          } else {
            debug_msg(1,
                      "options_load: wrong syntax of options line in %s line %d",
                      filename, i);
          }
        }
        g_strfreev(keyvalue);
      }
    }

    fclose(f);
  } else {
    debug_msg(1, "options_load: could not retrieve contents of file %s (%s)",
              filename, strerror(errno));
  }

  opt->filename = g_strdup(filename);
}
