#include <gtk/gtk.h>
#include <string.h>

#include "if_main.h"
#include "if_mcpconsole.h"
#include "world.h"
#include "debug.h"
#include "support.h"

#define MCP_WIDTH 500
#define MCP_HEIGHT 400

static GtkWidget *wndMcpConsole = NULL;
static GtkWidget *notebook_mcpConsole;
static GtkTextTagTable *mcpTagTable;

gboolean on_wndMcpConsole_delete(GtkWidget *widget, GdkEvent *event,
                                  gpointer user_data);
gboolean on_world_destroy(GtkWidget *widget, world *wld);

GtkWidget *
if_mcpconsole_create_page(world *wld) {
  GtkWidget *txt;
  GtkWidget *srl = gtk_scrolled_window_new(NULL, NULL);
  GdkColor bg;
  PangoFontDescription *f;
  
  gtk_widget_show(srl);
  gtk_container_set_border_width(GTK_CONTAINER(srl), 3);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(srl), GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(srl), GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);

  txt = gtk_text_view_new_with_buffer(wld->bufMcp);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(txt), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(txt), GTK_WRAP_WORD_CHAR);
  gtk_container_set_border_width(GTK_CONTAINER(txt), 3);
  gtk_container_add(GTK_CONTAINER(srl), txt);
  gtk_widget_show(txt);
    
  gdk_color_parse("black", &bg);
  gtk_widget_modify_base(txt, GTK_STATE_NORMAL, &bg);
  
  f =
    pango_font_description_from_string("monospace 8");
  
  if (f) {
    gtk_widget_modify_font(txt, f);
    pango_font_description_free(f);
  }
  
  g_signal_connect((gpointer)(wld->widgets.noteContainer), "destroy",
                   G_CALLBACK(on_world_destroy), wld);
  
  return srl;
}

void
if_mcpconsole_fill_notebook() {
  GList *wrlds;
  world *wld;
  
  for (wrlds = world_get_worlds(); wrlds; wrlds = wrlds->next) {
    wld = (world *)wrlds->data;
    
    if (wld->loaded) {
      gtk_notebook_append_page(GTK_NOTEBOOK(notebook_mcpConsole), 
                               if_mcpconsole_create_page(wld), gtk_label_new(
                               options_get_str(wld->settings, "name")));
    }
  }
}

GtkWidget *
if_mcpconsole_create() {
  if (wndMcpConsole != NULL) {
    return wndMcpConsole;
  }
  
  wndMcpConsole = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  
  gtk_window_set_title(GTK_WINDOW(wndMcpConsole), "Mcp console");
  gtk_window_set_transient_for(GTK_WINDOW(wndMcpConsole), 
                               GTK_WINDOW(if_main_get_widget("wndMain")));
  gtk_window_set_destroy_with_parent(GTK_WINDOW(wndMcpConsole), TRUE);
  gtk_window_set_skip_taskbar_hint(GTK_WINDOW(wndMcpConsole), TRUE);
  gtk_window_set_skip_pager_hint(GTK_WINDOW(wndMcpConsole), TRUE);
  
  gtk_window_set_default_size(GTK_WINDOW(wndMcpConsole), MCP_WIDTH, MCP_HEIGHT);
  gtk_container_set_border_width(GTK_CONTAINER(wndMcpConsole), 6);
  
  notebook_mcpConsole = gtk_notebook_new();
  gtk_widget_show(notebook_mcpConsole);
  gtk_container_set_border_width(GTK_CONTAINER(notebook_mcpConsole), 6);
  
  gtk_container_add(GTK_CONTAINER(wndMcpConsole), notebook_mcpConsole);
  
  if_mcpconsole_fill_notebook();
  
  g_signal_connect((gpointer)wndMcpConsole, "delete-event", 
                   G_CALLBACK(on_wndMcpConsole_delete), NULL);
  
  gtk_widget_show(wndMcpConsole);
  return wndMcpConsole;
}

GtkTextTag *
mcpconsole_create_tag(gchar *name, gchar *fg) {
  GtkTextTag *tag;
  GdkColor col;
  
  gdk_color_parse(fg, &col);
  tag = gtk_text_tag_new(name);
  g_object_set(G_OBJECT(tag), "foreground-gdk", &col, NULL);
  
  return tag;
}

void 
mcpconsole_init() {
  
  mcpTagTable = gtk_text_tag_table_new();
  gtk_text_tag_table_add(mcpTagTable, mcpconsole_create_tag("mcp-in", 
                                                            "#3EF73EF7BFFF"));
  gtk_text_tag_table_add(mcpTagTable, mcpconsole_create_tag("mcp-out", 
                                                            "#86318631FFFF"));
  gtk_text_tag_table_add(mcpTagTable, mcpconsole_create_tag("mcp-msg", 
                                                            "#0000A5290000"));
}

void 
mcpconsole_fini() {
  g_object_unref(mcpTagTable);
}

GtkTextTagTable *
if_mcpconsole_get_tag_table() {
  return mcpTagTable;
}

void mcpconsole_add(world *wld, mcpc_type mtype, gchar *msg) {
  gchar *m, *tagName = NULL, *newLine;
  gchar p[2];
  GtkTextIter end;
    
  p[1] = '\0';
  
  if (wld->loaded) {
    switch (mtype) {
      case MCPC_IN:
        p[0] = '<';
        tagName = g_strdup("mcp-in");
        break;
      case MCPC_OUT:
        p[0] = '>'; 
        tagName = g_strdup("mcp-out");
        break;
      case MCPC_MSG:
        p[0] = '#'; 
        tagName = g_strdup("mcp-msg");
        break;
    }
    
    m = g_strconcat(p, msg, "\n", NULL);
    gtk_text_buffer_get_end_iter(wld->bufMcp, &end);
    
    // convert to UTF-8
    newLine =
      g_convert_with_fallback(m, strlen(m), "UTF-8",
                              options_get_str(wld->settings, "charset"),
                              NULL, NULL, NULL, NULL);

    if (newLine == NULL) {  
      gtk_text_buffer_insert_with_tags_by_name(wld->bufMcp, &end, 
                                     m, strlen(m), tagName, NULL);
    } else {
      gtk_text_buffer_insert_with_tags_by_name(wld->bufMcp, &end, 
                                     newLine, strlen(newLine), tagName, NULL);
      g_free(newLine);
    }
    
    g_free(m);
    g_free(tagName);
  }
}

void
mcpconsole_add_world(world *wld) {
  if (wndMcpConsole != NULL) {
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook_mcpConsole), 
                             if_mcpconsole_create_page(wld), gtk_label_new(
                             options_get_str(wld->settings, "name")));
  }
}

/* CALLBACKS */

gboolean 
on_wndMcpConsole_delete(GtkWidget *widget, GdkEvent *event,
                         gpointer user_data) {
  wndMcpConsole = NULL;
  notebook_mcpConsole = NULL;
  
  return FALSE;
}

gboolean 
on_world_destroy(GtkWidget *widget, world *wld) {
  GtkNotebook *n;
  GtkWidget *p, *l;
  int r;
  int i;
  gchar *wname;
  
  if (wndMcpConsole == NULL) {
    return FALSE;
  }
  
  wname = options_get_str(wld->settings, "name");
  n = GTK_NOTEBOOK(notebook_mcpConsole);
  r = gtk_notebook_get_n_pages(n);
  
  for (i = 0; i < r; i++) {
    p = gtk_notebook_get_nth_page(n, i);
    l = gtk_notebook_get_tab_label(n, p);
    
    if (strcmp(wname, gtk_label_get_text(GTK_LABEL(l))) == 0) {
      gtk_widget_destroy(p);
      break;
    }
  }
  
  return FALSE;
}
