
#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>

#include "mcp.h"
#include "net.h"
#include "world.h"
#include "debug.h"
#include "g_options.h"
#include "packages.h"
#include "if_mcpconsole.h"

GList *registry = NULL;
GList *sessions = NULL;

GList *
mcp_get_registry() {
  return registry;
}

#define MCP_MIN_VERSION 2.1
#define MCP_MAX_VERSION 2.1

double
mcp_get_version(double clientMin, double clientMax, double serverMin,
                double serverMax) {
  if (clientMax >= serverMin && serverMax >= clientMin) {
    if (clientMax < serverMax) {
      return serverMax;
    } else {
      return clientMax;
    }
  } else {
    return 0.0;
  }
}

gchar *
mcp_find_value(GList * fields, gchar * key) {
  keyvaluepair *tmp;

  for (; fields; fields = fields->next) {
    tmp = (keyvaluepair *) (fields->data);
    if (strcasecmp(key, tmp->key) == 0) {
      return tmp->value;
    }
  }

  return NULL;
}

gchar *
mcp_escape_if_needed(gchar * line) {
  gchar *newLine;
  int i, j = 1;

  if (strlen(line) == 0) {
    return g_strdup("\"\"");
  }

  if (strchr(line, ' ') || strchr(line, '"') || strchr(line, '\\')) {
    newLine = g_new(gchar, (strlen(line) * 2) + 3);
    newLine[0] = '"';

    for (i = 0; i < strlen(line); i++) {
      if (line[i] == '"' || line[i] == '\\') {
        newLine[j] = '\\';
        j++;
      }
      newLine[j] = line[i];
      j++;
    }

    newLine[j] = '"';
    newLine[j + 1] = '\0';
    return newLine;
  } else {
    return g_strdup(line);
  }
}

gchar *
mcp_un_escape(gchar * line) {
  int i, j = 0;
  
  for (i = 0; i < strlen(line); i++) {
    if (line[i] == '\\') {
      i++;                      // Fast forward
    }

    line[j] = line[i];
    j++;
  }

  line[j] = '\0';
  return line;
}

gchar *
process_keyval(GList ** info, gchar * line) {
  int i = 0;
  gchar *keystart, *valuestart;
  int keylen, valuelen;
  keyvaluepair *newKeyValue;

  if (line[0] != ' ') {         // Must start with a space
    return NULL;
  }

  if (strlen(line) <= 1 || line[1] == ' ') {
    return NULL;
  }

  i = 1;

  while (i < strlen(line) && line[i] != ':' && line[i] != ' ') {
    i++;
  }

  if (i >= strlen(line) || line[i] != ':') {
    return NULL;
  }

  keystart = line + 1;
  keylen = i - 1;

  line = line + i + 1;
  i = 0;

  if ((i + 1) >= strlen(line) || line[i] != ' ') {
    return FALSE;
  }

  i++;

  if (line[i] == ' ') {
    return NULL;
  }

  valuestart = line + i;
  if (line[i] == '"') {
    valuestart++;
    i++;
    while (i < strlen(line) && line[i] != '"') {
      if (line[i] == '\\') {
        i++;
      }
      i++;
    }

    valuelen = (line + i) - valuestart;

    if (line[i] != '"') {
      valuelen--;
    } else {
      i++;
    }

  } else {
    while (i < strlen(line) && line[i] != ' ') {
      i++;
    }

    valuelen = (line + i) - valuestart;
  }

  newKeyValue = g_new(keyvaluepair, 1);
  newKeyValue->key = g_strndup(keystart, keylen);
  newKeyValue->value = mcp_un_escape(g_strndup(valuestart, valuelen));

  line = line + i;
  *info = g_list_append(*info, newKeyValue);

  return line;
}

GList *
process_key_values(gchar * line) {
  GList *result = NULL;

  while (line && strlen(line)) {
    line = process_keyval(&result, line);
  }

  return result;
}

gboolean
mcp_parse_line(gchar * line, message_info * info) {
  // We have a line, there should always be a name, so we'll do that first
  gchar *p;

  p = strchr(line, ' ');

  if (!p || p == line) {
    return FALSE;
  }

  info->name = g_strndup(line, (p - line));
  line = p + 1;

  // Now there should be a authentication key!
  p = strchr(line, ' ');

  if (!p) {
    p = line + strlen(line);
  }

  info->authkey = g_strndup(line, (p - line));
  line = p;

  // Now process the keyvals

  info->fields = process_key_values(line);
  return TRUE;
}

gchar *
mcp_generate_key(int len) {
  gchar *s = g_malloc(len + 1);
  gchar ref[] =
    "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_";
  int n = strlen(ref);
  int i;

  srand((int) time(NULL));

  for (i = 0; i < len; i++) {
    s[i] = ref[rand() % n];
  }

  s[i] = 0;
  return s;
}

gchar *
mcp_generate_data_tag() {
  return mcp_generate_key(10);
}

gchar *
mcp_generate_auth_key() {
  return mcp_generate_key(6);
}

void
mcp_send_multiline(mcp_session * session, gchar * data_tag, gchar * key,
                   GList * lines) {
  gchar *msg = g_strconcat("#$#* ", data_tag, " ", key, ": ", NULL);
  gchar *msgLine = NULL;

  for (; lines; lines = lines->next) {
    msgLine = g_strconcat(msg, (gchar *) (lines->data), NULL);
    net_send_line(session->owner, msgLine);
    
    world_log(session->owner, LOG_MCP_OUT, msgLine);    
    mcpconsole_add(session->owner, MCPC_OUT, msgLine);
    
    g_free(msgLine);
  }

  g_free(msg);
  msgLine = g_strconcat("#$#: ", data_tag, NULL);
  net_send_line(session->owner, msgLine);
  
  world_log(session->owner, LOG_MCP_OUT, msgLine);
  mcpconsole_add(session->owner, MCPC_OUT, msgLine);
  
  g_free(msgLine);
}

void
mcp_send_simple(mcp_session * session, gchar * pname, gchar * firstKey, ...) {
  GString *msg = NULL;
  va_list args;
  gchar *key, *value, *esvalue;

  msg = g_string_new("#$#");
  msg = g_string_append(msg, pname);

  if (strcasecmp(pname, "mcp") != 0) {
    msg = g_string_append(msg, " ");
    msg = g_string_append(msg, session->authkey);
  }

  msg = g_string_append(msg, " ");

  if (firstKey != NULL) {
    va_start(args, firstKey);
    key = firstKey;

    while (key) {
      value = va_arg(args, gchar *);

      if (value) {
        msg = g_string_append(msg, key);
        msg = g_string_append(msg, ": ");
        esvalue = mcp_escape_if_needed(value);
        msg = g_string_append(msg, esvalue);
        g_free(esvalue);
        msg = g_string_append(msg, " ");
        key = va_arg(args, gchar *);
      } else {
        key = NULL;
      }
    }

    va_end(args);
  }

  net_send_line(session->owner, msg->str);
  
  world_log(session->owner, LOG_MCP_OUT, msg->str);
  mcpconsole_add(session->owner, MCPC_OUT, msg->str);
  
  g_string_free(msg, TRUE);
}

void
mcp_destroy_fields(GList * fields) {
  keyvaluepair *tmp;

  for (; fields; fields = fields->next) {
    tmp = (keyvaluepair *) (fields->data);
    g_free(tmp->key);
    g_free(tmp->value);
    g_free(fields->data);
    fields->data = NULL;
  }

  g_list_free(fields);
}

package_registration *
mcp_find_package(gchar * pname) {
  GList *elem;
  package_registration *p;

  for (elem = registry; elem; elem = elem->next) {
    p = (package_registration *) (elem->data);
    if (strncasecmp(pname, p->name, strlen(p->name)) == 0
        && (strlen(pname) == strlen(p->name)
            || pname[strlen(p->name)] == '-')) {
      return p;
    }
  }

  return NULL;
}

mcp_package *
mcp_can_handle(mcp_session * session, package_registration * reg) {
  GList *elem;
  mcp_package *m;

  for (elem = session->packages; elem; elem = elem->next) {
    m = (mcp_package *) (elem->data);
    if (m->registration == reg) {
      return m;
    }
  }

  return NULL;
}

mcp_session *
mcp_find_session(world * wld) {
  GList *elem;
  mcp_session *session;

  for (elem = sessions; elem; elem = elem->next) {
    session = (mcp_session *) (elem->data);
    if (session->owner == wld) {
      return session;
    }
  }

  return NULL;
}

mcp_package *
mcp_create_package_for(mcp_session * session, package_registration * p) {
  mcp_package *newPackage;

  if (!p) {
    return NULL;
  }

  if (!mcp_can_handle(session, p)) {
    newPackage = g_new(mcp_package, 1);
    newPackage->registration = p;
    newPackage->session = session;
    newPackage->user_data = NULL;
    newPackage->multiline = NULL;
    session->packages = g_list_append(session->packages, newPackage);
    return newPackage;
  } else {
    debug_msg(2,
              "mcp_create_package_for: package (%s) is already registered for %s",
              p->name, options_get_str(session->owner->settings, "name"));
    return NULL;
  }

}

void
mcp_handle_open(world * wld, gchar * line) {
  GList *fields = process_key_values(line);
  mcp_session *newSession;
  mcp_package *p;

  gchar *minVersion = mcp_find_value(fields, "version");
  gchar *maxVersion = mcp_find_value(fields, "to");
  double minV, maxV, version;

  if (!(minVersion && maxVersion)) {
    debug_msg(2,
              "mcp_handle_open: invalid opening, minVersion OR maxVersion not found!");
    mcpconsole_add(wld, MCPC_MSG, "invalid open of session, 'version' or 'to' key not found");
    return;
  }

  fix_decimal_point(minVersion, strlen(minVersion));
  fix_decimal_point(maxVersion, strlen(maxVersion));
  
  minV = strtod(minVersion, NULL);
  maxV = strtod(maxVersion, NULL);

  version = mcp_get_version(MCP_MIN_VERSION, MCP_MAX_VERSION, minV, maxV);

  if (version > 0.0) {
    newSession = g_new(mcp_session, 1);
    newSession->authkey = mcp_generate_auth_key();
    newSession->version = version;
    newSession->packages = NULL;
    newSession->owner = wld;

    sessions = g_list_append(sessions, newSession);
    mcp_send_simple(newSession, "mcp", "authentication-key",
                    newSession->authkey, "version", "2.1", "to", "2.1", NULL);

    p = mcp_create_package_for(newSession, mcp_find_package("mcp-negotiate"));
    p->version = 1.0;

    if (p->registration->initFunc) {
      (p->registration->initFunc) (p);
    }
  } else {
    debug_msg(2,
              "mcp_handle_open: mcp server version does NOT match client version!");
    mcpconsole_add(wld, MCPC_MSG, "mcp server version does NOT match client version");
  }
}

gchar *
mcp_find_multiline_tag(GList * fields) {
  keyvaluepair *data;

  for (; fields; fields = fields->next) {
    data = (keyvaluepair *) (fields->data);
    if (data->key[strlen(data->key) - 1] == '*') {
      return data->key;
    }
  }

  return NULL;
}

void
mcp_invoke_multiline_handle(mcp_session * session, gchar * data_tag,
                            gchar * key, gchar * value) {
  GList *t;
  GList *m;
  GList *elem;
  mcp_package *p;
  mcp_multiline_info *minfo;
  gboolean handleAll = FALSE;

  // Now find the data_tag and emit the handle_multiline
  for (t = session->packages; t; t = t->next) {
    p = (mcp_package *) (t->data);
    for (m = p->multiline; m; m = m->next) {
      minfo = (mcp_multiline_info *) (m->data);
      if (strcmp(minfo->data_tag, data_tag) == 0) {
        if (p->registration->multiFunc) {
          if (key) {
            if (!(p->registration->multiFunc(p, data_tag, key, value, NULL))) {
              // Not handled a single value, store it for later
              minfo->data = g_list_append(minfo->data, g_strdup(value));
            }
          } else {
            handleAll =
              p->registration->multiFunc(p, data_tag, NULL, NULL, minfo->data);
          }
        }

        if (!key) {
          // This is the end, remove the minfo
          g_free(minfo->data_tag);
          g_free(minfo->key);

          // Lines should only be freed when allValues is not handled!
          if (!handleAll) {
            for (elem = minfo->data; elem; elem = elem->next) {
              g_free((gchar *) (elem->data));
            }
          }

          g_list_free(minfo->data);

          p->multiline = g_list_remove(p->multiline, minfo);
          g_free(minfo);
        }
        return;
      }
    }
  }
}

void
mcp_handle_multiline(mcp_session * session, gchar * line) {
  // parse the multiline
  int i = 0, keyStart;
  gchar *data_tag, *key, *value;

  while (line[i] != '\0' && line[i] != ' ') {
    i++;
  }

  if (line[i] == '\0') {
    return;
  }

  data_tag = g_strndup(line, i);

  i++;
  keyStart = i;

  while (line[i] != '\0' && line[i] != ' ') {
    i++;
  }

  if (line[i] == '\0') {
    return;
  }

  key = g_strndup(line + keyStart, (i - keyStart) - 1);
  value = g_strdup(line + i + 1);
  mcp_invoke_multiline_handle(session, data_tag, key, value);

  g_free(data_tag);
  g_free(key);
  g_free(value);
}

void
mcp_handle_multiline_end(mcp_session * session, gchar * line) {
  int i = 0;
  gchar *data_tag;

  while (line[i] != '\0' && line[i] != ' ') {
    i++;
  }

  data_tag = g_strndup(line, i);
  mcp_invoke_multiline_handle(session, data_tag, NULL, NULL);
  g_free(data_tag);
}

void
mcp_handle_out_of_bound(world * wld, gchar * line) {
  message_info info;
  mcp_session *session = mcp_find_session(wld);
  package_registration *p;
  mcp_package *m;
  mcp_multiline_info *minfo;
  gchar *suffix, *value;
  gchar *full;
  
  debug_msg(2, "mcp_handle_out_of_bound: %s", line);
  
  full = g_strconcat("#$#", line, NULL);
  mcpconsole_add(wld, MCPC_IN, full);
  g_free(full);

  info.fields = NULL;
  info.authkey = NULL;
  info.name = NULL;

  if (strncmp(line, "mcp ", 4) == 0 && !session) {
    mcp_handle_open(wld, line + 3);
  } else if (!session) {
    // No session, drop
    debug_msg(2,
              "mcp_handle_out_of_bound: no session found for %s, ignoring out of bound command!",
              options_get_str(wld->settings, "name"));
    mcpconsole_add(wld, MCPC_MSG, "no session is found, ignoring the message");
  } else if (strncmp(line, "* ", 2) == 0) { // Multiline
    mcp_handle_multiline(session, line + 2);
  } else if (strncmp(line, ": ", 2) == 0) { // Multiline end
    mcp_handle_multiline_end(session, line + 2);
  } else {
    if (mcp_parse_line(line, &info)) {
      if (strcmp(info.authkey, session->authkey) == 0) {
        if ((p = mcp_find_package(info.name))
            && (m = mcp_can_handle(session, p))) {
          if ((value = mcp_find_multiline_tag(info.fields))) {
            if (mcp_find_value(info.fields, "_data-tag")) {
              // This package requests a multiline opening!
              minfo = g_new(mcp_multiline_info, 1);
              minfo->key = g_strdup(value);
              minfo->data_tag =
                g_strdup(mcp_find_value(info.fields, "_data-tag"));
              minfo->data = NULL;
              m->multiline = g_list_append(m->multiline, minfo);
              debug_msg(2,
                        "mcp_handle_out_of_bound: multiline opening detected and stored (data_tag = %s, key = %s)",
                        minfo->data_tag, minfo->key);
            } else {
              debug_msg(2,
                        "mcp_handle_out_of_bound: multiline opening detected, but no _data-tag specified, ignore message");
              mcpconsole_add(wld, MCPC_MSG, "no _data-tag specified, ignoring the message");
            }
          }

          if (p->simpleFunc) {
            if (strlen(p->name) < strlen(info.name)) {
              suffix = info.name + (strlen(p->name) + 1);
            } else {
              suffix = NULL;
            }

            p->simpleFunc(m, suffix, info.fields);
          }
        } else {
          debug_msg(2,
                    "mcp_handle_out_of_bound: unsupported package %s, ignore message",
                    info.name);
          mcpconsole_add(wld, MCPC_MSG, "package is not supported, ignore the message");
        }
      } else {
        debug_msg(2,
                  "mcp_handle_out_of_bound: invalid authentication key %s instead of %s, ignore message",
                  info.authkey, session->authkey);
        mcpconsole_add(wld, MCPC_MSG, "invalid authentication, ignore the message");
      }
    } else {
      debug_msg(2,
                "mcp_handle_out_of_bound: invalid message (could not parse), ignore message");
      mcpconsole_add(wld, MCPC_MSG, "invalid message (parse error), ignore the message");
    }
    // Free the info
    if (info.authkey) {
      g_free(info.authkey);
    }
    if (info.name) {
      g_free(info.name);
    }
    if (info.fields) {
      mcp_destroy_fields(info.fields);
    }
  }
}

GList *
mcp_find_packages(gchar *packages) {
  GList *result = NULL;
  gchar **p;
  int i;
  package_registration *reg;
  
  p = g_strsplit(packages, ", ", -1);
  
  for (i = 0; p[i]; i++) {
    if ((reg = mcp_find_package(p[i]))) {
      result = g_list_append(result, reg);  
    }  
  }
  
  g_strfreev(p);
  return result;
}

mcp_package *
mcp_session_find_package(mcp_session *session, gchar *name) {
  GList *pack;
  mcp_package *p;
  
  for (pack = session->packages; pack; pack = pack->next) {
    p = (mcp_package *)(pack->data);
    
    if (strcmp(p->registration->name, name) == 0) {
      return p;
    }
  }
  
  return NULL;
}

void
mcp_register_package(gchar * name, double minVersion, double maxVersion,
                     mcp_init_function initFunc,
                     mcp_handle_simple_function simpleFunc,
                     mcp_handle_multi_function multiFunc,
                     mcp_fini_function finiFunc, gchar *overrides, 
                     gchar *depends) {
  package_registration *reg = g_new(package_registration, 1);

  debug_msg(2, "mcp_register_package: registering package %s", name);

  reg->name = g_strdup(name);
  reg->minVersion = minVersion;
  reg->maxVersion = maxVersion;

  reg->initFunc = initFunc;
  reg->finiFunc = finiFunc;
  reg->simpleFunc = simpleFunc;
  reg->multiFunc = multiFunc;

  if (overrides) {
    reg->overrides = mcp_find_packages(overrides);
  } else {
    reg->overrides = NULL;
  }
  
  if (depends) {
    reg->depends = mcp_find_packages(depends);
  } else {
    reg->depends = NULL;
  }

  registry = g_list_append(registry, reg);
}

void
mcp_init() {
  packages_init();
}

void
mcp_destroy_session(mcp_session * session) {
  GList *p, *pp;
  GList *t;
  mcp_package *pi;
  mcp_multiline_info *minfo;
 
  mcpconsole_add(session->owner, MCPC_MSG, "destroying session");
  pp = g_list_copy(session->packages);
  
  for (p = pp; p; p = p->next) {
    pi = (mcp_package *) (p->data);

    if (pi->registration && pi->registration->finiFunc) {
      // finiFunc should take care of user_data
      debug_msg(2, "mcp_destroy_session: finalizing %s",
                pi->registration->name);
      pi->registration->finiFunc(pi);
    }
    // Free'ing outstanding multilines
    for (t = pi->multiline; t; t = t->next) {
      minfo = (mcp_multiline_info *) (t->data);
      g_free(minfo->key);
      g_free(minfo->data_tag);
      g_free(minfo);
    }

    g_list_free(pi->multiline);
    
    session->packages = g_list_remove(session->packages, pi);
    g_free(pi);
  }

  g_list_free(pp);
  g_list_free(session->packages);
  session->packages = NULL;
  g_free(session->authkey);

  sessions = g_list_remove(sessions, session);
  g_free(session);
}

// Clear all, session SHOULD be already cleared, but you can't be really sure
void
mcp_fini() {
  GList *p;
  package_registration *pi;
  mcp_session *session;

  // Sessions if there are any??
  for (p = sessions; p; p = sessions) { // Because mcp_destroy_session also removes the session from the list
    session = (mcp_session *) (p->data);
    debug_msg(2, "mcp_fini: destroying session for %s",
              options_get_str(session->owner->settings, "name"));
    mcp_destroy_session(session);
  }

  g_list_free(sessions);
  sessions = NULL;

  for (p = registry; p; p = p->next) {
    pi = (package_registration *) (p->data);
    
    g_free(pi->name);
    g_list_free(pi->overrides);
    g_list_free(pi->depends);

    g_free(pi);
  }

  g_list_free(registry);
  registry = NULL;
}

void
mcp_close(world * wld) {
  // set all negotiating to FALSE, call fini functions of negotiated packages
  mcp_session *session = mcp_find_session(wld);

  if (session) {
    mcp_destroy_session(session);
  } else {
    debug_msg(2, "mcp_close: no session for %s, can't close!",
              options_get_str(wld->settings, "name"));
  }
}
