#include <gtk/gtk.h>
#include <glib.h>
#include <string.h>
#include <glob.h>
#include <unistd.h>
#include <errno.h>

#include "debug.h"
#include "mcp.h"
#include "packages.h"
#include "list.h"
#include "my-string.h"
#include "support.h"
#include "world.h"
#include "gnoe_pixbuf.h"
#include "g_options.h"

#define ICE_UINFO(x) (ice_userlist_info *)(x->user_data)

typedef struct _ice_userlist_url_map {
  gchar *name;
  gchar *filename;
} ice_userlist_url_map;

typedef struct _ice_userlist_key_datatag_map {
  gchar *key;
  gchar *data_tag;
} ice_userlist_key_datatag_map;

static ice_userlist_url_map default_states[] = {
  {"away+idle", "ice-userlist/away+idle.svg"},
  {"busy+idle", "ice-userlist/busy+idle.svg"},
  {"away", "ice-userlist/away.svg"},
  {"busy", "ice-userlist/busy.svg"},
  {"avail+idle", "ice-userlist/avail+idle.svg"},
  {NULL, NULL}
};

static ice_userlist_url_map default_ranks[] = {
  {"inhabitant", "ice-userlist/inhabitant.svg"},
  {"programmer", "ice-userlist/programmer.svg"},
  {"wizard", "ice-userlist/wizard.svg"},
  {NULL, NULL}
};

typedef struct _ice_userlist_info {
  GList *state_icons;
  GList *rank_icons;
  
  GList *key_datatags;
  GnoeMoeFetchHandle *rank_fetch;
  GnoeMoeFetchHandle *state_fetch;
  
  gchar *rank_dir;
  gchar *state_dir;
  
} ice_userlist_info;

void ice_userlist_player_add(mcp_package *p, playerdb_player_info *ppi, 
                             gchar *key, gchar *value);
void ice_userlist_player_set(mcp_package *p, playerdb_player_info *ppi, 
                             gchar *key, gchar *value);
void ice_userlist_player_del(mcp_package *p, playerdb_player_info *ppi, 
                             gchar *key, gchar *value);

ice_userlist_url_map *
ice_userlist_url_map_new(gchar *name, gchar *filename) {
  ice_userlist_url_map *r = g_new(ice_userlist_url_map, 1);
  
  r->name = g_strdup(name);
  r->filename = g_strdup(filename);
  
  return r;
}

GList *
ice_userlist_find_icon_names(GList *icons, gchar *path) {
  GList *item, *res = NULL;
  ice_userlist_url_map *map;

  for (item = icons; item; item = item->next) {
    map = (ice_userlist_url_map *)(item->data);
    
    if (strcmp(map->filename, path) == 0) {
      res = g_list_append(res, map->name);
    }    
  }
  
  return res;
}

int
ice_userlist_find_path(GList *icons, gchar *name, gchar **path) {
  GList *item;
  ice_userlist_url_map *map;

  for (item = icons; item; item = item->next) {
    map = (ice_userlist_url_map *)(item->data);
    
    if (strcmp(map->name, name) == 0) {
      *path = map->filename;
      return g_list_position(icons, item);
    }    
  }
  
  *path = NULL;
  return g_list_length(icons) + 1;
}

int
ice_userlist_get_icon(GList *icons, gchar *name, GdkPixbuf **pix) {
  gchar *path;
  int ret;
  
  *pix = NULL;
  ret = ice_userlist_find_path(icons, name, &path);
  
  if (path) {
    *pix = gnoe_pixbuf_get(path);
  }
  
  return ret;
}

GList *
ice_userlist_remove_datatag(GList *items, gchar *datatag) {
  GList *item;
  ice_userlist_key_datatag_map *map;

  for (item = items; item; item = item->next) {
    map = (ice_userlist_key_datatag_map *)(item->data);
    
    if (strcmp(map->data_tag, datatag) == 0) {
      g_free(map->key);
      g_free(map->data_tag);
      g_free(map);
      return g_list_remove(items, map);
    }
  }
  
  return items;
}

gchar *
ice_userlist_find_key(GList *items, gchar *datatag) {
  GList *item;
  ice_userlist_key_datatag_map *map;

  for (item = items; item; item = item->next) {
    map = (ice_userlist_key_datatag_map *)(item->data);
    
    if (strcmp(map->data_tag, datatag) == 0) {
      return map->key;
    }    
  }
  
  return NULL;
}

void
ice_userlist_init_states(ice_userlist_info *i) {
  int n = 0;
  
  i->state_icons = NULL;
  
  while (!(default_states[n].name == NULL && default_states[n].filename == NULL)) {
    i->state_icons = g_list_append(i->state_icons, 
                          ice_userlist_url_map_new(default_states[n].name,
                             default_states[n].filename));
    n++;
  }
}

void
ice_userlist_init_ranks(ice_userlist_info *i) {
  int n = 0;
  
  i->rank_icons = NULL;
  
  while (!(default_ranks[n].name == NULL && default_ranks[n].filename == NULL)) {    
    i->rank_icons = g_list_append(i->rank_icons, 
                          ice_userlist_url_map_new(default_ranks[n].name,
                            default_ranks[n].filename));
    n++;
  }
}

void
mcp_ice_userlist_init(mcp_package * p) {
  ice_userlist_info *i = g_new(ice_userlist_info, 1);
  gchar *iconsDir;
  
  ice_userlist_init_states(i);  
  ice_userlist_init_ranks(i);  
  
  mcp_playerdb_signal_connect(p->session, SIG_ADD, NULL, 
                             ice_userlist_player_add);
  mcp_playerdb_signal_connect(p->session, SIG_SET, "P_NAME", 
                             ice_userlist_player_set);
  mcp_playerdb_signal_connect(p->session, SIG_SET, "P_RANK", 
                             ice_userlist_player_set);
  mcp_playerdb_signal_connect(p->session, SIG_SET, "P_STATE", 
                             ice_userlist_player_set);
  mcp_playerdb_signal_connect(p->session, SIG_DEL, NULL, 
                             ice_userlist_player_del);

  world_userlist_create(p->session->owner);
  i->key_datatags = NULL;
  i->rank_fetch = NULL;
  i->state_fetch = NULL;
  
  iconsDir = g_strconcat(p->session->owner->path, "/icons", NULL);
  i->rank_dir = g_strconcat(iconsDir, "/ranks", NULL);
  i->state_dir = g_strconcat(iconsDir, "/states", NULL);
  
  if (!g_file_test(iconsDir, G_FILE_TEST_EXISTS))
    mkdir(iconsDir, 0755);
  if (!g_file_test(i->rank_dir, G_FILE_TEST_EXISTS))
    mkdir(i->rank_dir, 0755);
  if (!g_file_test(i->state_dir, G_FILE_TEST_EXISTS))
    mkdir(i->state_dir, 0755);
    
  g_free(iconsDir);  
  p->user_data = (gpointer)i;
}

void
ice_userlist_free_key_datags(GList *rs) {
  GList *item;
  ice_userlist_key_datatag_map *map;

  for (item = rs; item; item = item->next) {
    map = (ice_userlist_key_datatag_map *)(item->data);
    
    g_free(map->key);
    g_free(map->data_tag);
    g_free(map);
  }
  
  g_list_free(rs);
}

void
ice_userlist_free_list(GList *rs) {
  GList *item;
  ice_userlist_url_map *map;

  for (item = rs; item; item = item->next) {
    map = (ice_userlist_url_map *)(item->data);
    
    g_free(map->name);
    g_free(map->filename);
    g_free(map);
  }
  
  g_list_free(rs);
}

void
mcp_ice_userlist_fini(mcp_package *p) {
  ice_userlist_info *i = ICE_UINFO(p);
  
  mcp_playerdb_signal_disconnect(p->session, ice_userlist_player_add);
  mcp_playerdb_signal_disconnect(p->session, ice_userlist_player_set);
  mcp_playerdb_signal_disconnect(p->session, ice_userlist_player_del);
  
  ice_userlist_free_list(i->state_icons);
  ice_userlist_free_list(i->rank_icons);
  ice_userlist_free_key_datags(i->key_datatags);
  
  world_userlist_free(p->session->owner);
  g_free(i->rank_dir);
  g_free(i->state_dir);
  
  if (i->rank_fetch) {
    i->rank_fetch->aborted = TRUE;
  }
  
  if (i->state_fetch) {
    i->state_fetch->aborted = TRUE;
  }
  
  g_free(i);
}

void
ice_userlist_update_player(mcp_package *p, playerdb_player_info *ppi) {
  ice_userlist_info *i = (ice_userlist_info *)(p->user_data);
  gchar *state;
  world_uu uuinfo;
  unsigned int sortid = 0;

  uuinfo.nr = ppi->id;
  
  /* Construct the name */
  uuinfo.name = g_strdup_printf("%s (#%d)", playerdb_get_prop(ppi, "P_NAME"), ppi->id);
  
  /* Find the appropriate icon */
  state = playerdb_get_prop(ppi, "P_STATE");
  
  if (strcmp(state, "avail") == 0) {
    sortid = g_list_length(i->rank_icons) - ice_userlist_get_icon(
               i->rank_icons, playerdb_get_prop(ppi, "P_RANK"), &(uuinfo.icon)) 
               - 1;
  } else {
    sortid = (g_list_length(i->state_icons) - ice_userlist_get_icon(
                i->state_icons, state, &(uuinfo.icon)) - 1) + 
                g_list_length(i->rank_icons);
  }

  uuinfo.sort =
    g_strdup_printf("%.2d%s", sortid, uuinfo.name);
  
  debug_msg(2, "Update player: %s", uuinfo.sort);
  world_userlist_update_user(p->session->owner, &uuinfo);

  g_free(uuinfo.name);
  g_free(uuinfo.sort);
}

void
ice_userlist_player_add(mcp_package *p, playerdb_player_info *ppi, 
                                     gchar *key, gchar *value) {
  debug_msg(2, "ice_userlist_player_add: %d", ppi->id);
  ice_userlist_update_player(
    mcp_session_find_package(p->session, "dns-nl-icecrew-userlist"), ppi);
}

void
ice_userlist_player_set(mcp_package *p, playerdb_player_info *ppi, 
                                     gchar *key, gchar *value) {
  debug_msg(2, "ice_userlist_player_set: %d, %s to %s", ppi->id, key, value);
  ice_userlist_update_player(
    mcp_session_find_package(p->session, "dns-nl-icecrew-userlist"), ppi);
}

void
ice_userlist_player_del(mcp_package *p, playerdb_player_info *ppi, 
                                     gchar *key, gchar *value) {
  debug_msg(2, "ice_userlist_player_del: %d", ppi->id);
  world_userlist_remove_user(p->session->owner, ppi->id);
}

gchar *
ice_userlist_find_name(gchar *name, ice_userlist_url_map def[]) {
  int i = 0;
  
  while (!(def[i].name == NULL && def[i].filename == NULL)) {
    if (strcmp(def[i].name, name) == 0) {
      return def[i].filename;
    }
    
    i++;
  } 
  
  return NULL; 
}

void
ice_userlist_remove_not_used_icons(gchar *dir, GList *used) {
  glob_t gl;
  gchar *g = g_strconcat(dir, "/*", NULL);
  int i;
  
  glob(g, 0, NULL, &gl);
  
  for (i = 0; i < gl.gl_pathc; i++) {
    if (g_list_find_simple(used, gl.gl_pathv[i]) == NULL) {
      debug_msg(2, "ice_userlist_remove_not_used_icons: removing %s", gl.gl_pathv[i]);
      if (unlink(gl.gl_pathv[i]) == -1) {
        debug_msg(2, "ice_userlist_remove_not_used_icons: error %s", strerror(errno));
      }
    }
  }
  
  g_free(g);
  globfree(&gl);
}

void
ice_userlist_update_rank_icon(playerdb_player_info *ppi, gpointer user_data) {
  mcp_package *p = (mcp_package *)(user_data);
  gchar *state;
  
  state = playerdb_get_prop(ppi, "P_STATE");
  debug_msg(2, "ice_userlist_update_rank_icon: %s", state);
  
  if (state && strcmp(state, "avail") == 0) {
    ice_userlist_update_player(p, ppi);
  }
}

void
ice_userlist_update_state_icon(playerdb_player_info *ppi, gpointer user_data) {
  mcp_package *p = (mcp_package *)(user_data);  
  ice_userlist_update_player(p, ppi);
}

void 
ice_userlist_new_rank_state_icon(mcp_package *p, gchar *path, gboolean rank) {
  ice_userlist_info *i = ICE_UINFO(p);
  GList *names;
  GList *tmp;
  
  if (rank) {
    names = ice_userlist_find_icon_names(i->rank_icons, path);
  } else {
    names = ice_userlist_find_icon_names(i->state_icons, path);
  }
  
  if (names == NULL) {
    return;
  }
  
  for (tmp = names; tmp; tmp = tmp->next) {
    if (rank) {
      playerdb_find_players_with(p->session, "P_RANK", (gchar *)(tmp->data), 
                                 ice_userlist_update_rank_icon, p);
    } else {
      playerdb_find_players_with(p->session, "P_STATE", (gchar *)(tmp->data), 
                               ice_userlist_update_state_icon, p);    
    }
  }
  
  g_list_free(names);
}

void
ice_userlist_fetch_progress(GnoeMoeFetchHandle *g, mcp_package *p) {
  ice_userlist_info *i = ICE_UINFO(p);
  char *path;
  
  if (g->aborted) {
    if (i->rank_fetch == g) {
      i->rank_fetch = NULL;
    } else {
      i->state_fetch = NULL;
    }
    
    debug_msg(2, "ice_userlist_fetch_progress: fetch aborted!");
    return;
  }
  
  if (g->done) {
    if (i->rank_fetch == g) {
      i->rank_fetch = NULL;
    } else {
      i->state_fetch = NULL;
    }
    
    debug_msg(2, "ice_userlist_fetch_progress: fetch done!");
    return;
  }
  
  if (g->cur_phase == GNOME_VFS_XFER_PHASE_FILECOMPLETED) {
    path = gnome_vfs_get_local_path_from_uri(g->cur_file_name);
    debug_msg(2, "ice_userlist_fetch_progress: %s fetched!", path);

    if (i->rank_fetch == g) {
      ice_userlist_new_rank_state_icon(p, path, TRUE);
    } else {
      ice_userlist_new_rank_state_icon(p, path, FALSE);
    }
    
    g_free(path);
  }
}

void
ice_userlist_handle_ranks_states(mcp_package *p, GList *values, gboolean rank) {
  ice_userlist_info *i = ICE_UINFO(p);
  GList *tmp, *data, *newList = NULL, *source_uri = NULL, *dest_uri = NULL;
  gchar *name, *path, *url, *base;
  int len;
  
  for (tmp = values; tmp; tmp = tmp->next) {
    data = (GList *)(tmp->data);
    len = g_list_length(data);
    
    if (len == 0) {
      continue;
    }
    
    name = g_list_nth_data(data, 0);
    ice_userlist_find_path(newList, name, &path);
    
    if (path) {
      debug_msg(2, "ice_userlist_handle_ranks_states: duplicate %s, ignored!", name);
      continue;
    }
    
    if (len == 1) {
      if (rank) {
        path = ice_userlist_find_name(name, default_ranks);
      } else {
        path = ice_userlist_find_name(name, default_states);
      }
      
      if (path) {
        newList = g_list_append(newList, 
                      ice_userlist_url_map_new(name, path));
      } else {
        debug_msg(2, "ice_userlist_handle_ranks_states: couldn't find %s, ignored!", name);
      }
    } else {
      url = g_list_nth_data(data, 1);
      base = g_path_get_basename(url);
      
      if (base && strlen(base) > 1) {
        if (rank) {
          path = g_strconcat(i->rank_dir, "/", base, NULL);
        } else {
          path = g_strconcat(i->state_dir, "/", base, NULL);
        }
        
        source_uri = g_list_append(source_uri, strdup(url));
        dest_uri = g_list_append(dest_uri, strdup(path));
        newList = g_list_append(newList, 
                                 ice_userlist_url_map_new(name, path));
        g_free(path);
      }
      
      g_free(base);
    }
  }

  if (rank) {
    ice_userlist_free_list(i->rank_icons);
    i->rank_icons = newList;
  } else {
    ice_userlist_free_list(i->state_icons);
    i->state_icons = newList;  
  }
  
  if (rank && i->rank_fetch) {
    i->rank_fetch->aborted = TRUE;
  } else if (!rank && i->state_fetch) {
    i->state_fetch->aborted = TRUE;
  }
  
  while ((rank && i->rank_fetch != NULL) || (!rank && i->state_fetch != NULL)) {
    doEvents();
  }
  
  if (rank) {
    i->rank_fetch = gnoemoe_fetch(source_uri, dest_uri, 
                                (GFunc)ice_userlist_fetch_progress, p);
    ice_userlist_remove_not_used_icons(i->rank_dir, dest_uri);
  } else {
    i->state_fetch = gnoemoe_fetch(source_uri, dest_uri, 
                                (GFunc)ice_userlist_fetch_progress, p);  
    ice_userlist_remove_not_used_icons(i->state_dir, dest_uri);
  }
  
  
  g_list_free_simple(source_uri);
  g_list_free_simple(dest_uri);
}

gboolean 
mcp_ice_userlist_multi(mcp_package * p, gchar * data_tag, gchar * key,
                       gchar * value, GList * allValues) {
  ice_userlist_info *i = ICE_UINFO(p);
  GList *tmp, *l, *vals = NULL;
  ice_userlist_key_datatag_map *m;
  
  if (key) {
    if (!ice_userlist_find_key(i->key_datatags, data_tag)) {
      m = g_new(ice_userlist_key_datatag_map, 1);
      m->key = g_strdup(key);
      m->data_tag = g_strdup(data_tag);
      
      i->key_datatags = g_list_append(i->key_datatags, m);
    }
  } else if ((key = ice_userlist_find_key(i->key_datatags, data_tag)) != NULL) {
    for (tmp = allValues; tmp; tmp = tmp->next) {
      l = mcp_parse_list((gchar *)(tmp->data));
      vals = g_list_append(vals, l);
    }
    
    if (strcmp(key, "rank") == 0) {      
      ice_userlist_handle_ranks_states(p, vals, TRUE);
    } else if (strcmp(key, "state") == 0) {
      ice_userlist_handle_ranks_states(p, vals, FALSE);
    }
    
    i->key_datatags = ice_userlist_remove_datatag(i->key_datatags, data_tag);
  }
  
  for (tmp = vals; tmp; tmp = tmp->next) {
    mcp_free_list((GList *)(tmp->data));
  }
  
  g_list_free(vals);
  return FALSE;
}
