#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "mcp.h"
#include "debug.h"
#include "if_mcpconsole.h"

typedef struct _negotiate_package_info {
  package_registration *reg;
  double version;
} negotiate_package_info;

typedef struct _negotiate_info {
  GList *packages;
} negotiate_info;

void
mcp_negotiate_init(mcp_package * p) {
  negotiate_info *ni = g_new(negotiate_info, 1);
  
  ni->packages = NULL;
  p->user_data = ni;  
}

void
mcp_negotiate_fini(mcp_package * p) {
  negotiate_info *ni = (negotiate_info *)(p->user_data);
  GList *l;
  
  if (ni->packages) {
    for (l = ni->packages; l; l = l->next) {
      g_free(l->data);
    }
    
    g_list_free(ni->packages);
  }
  
  g_free(ni);
}

negotiate_package_info *
negotiate_find_package(GList *packages, package_registration *reg) {
  negotiate_package_info *npi;
  
  for ( ;packages; packages = packages->next) {
    npi = (negotiate_package_info *)(packages->data);
    
    if (npi->reg == reg) {
      return npi;
    }  
  }
  
  return NULL;
}

GList *
negotiate_fix_overrides(GList *packages) {
  negotiate_package_info *npi, *npii;
  GList *l, *item, *over;
  package_registration *reg;
  
  l = g_list_copy(packages);
  for (item = l; item; item = item->next) {
    npi = (negotiate_package_info *)(item->data);
    
    for (over = npi->reg->overrides; over; over = over->next) {
      reg = (package_registration *)(over->data);
      
      if ((npii = negotiate_find_package(packages, reg))) {
        debug_msg(2, "negotiate_fix_overrides: package %s overrides %s",
        npi->reg->name, reg->name);
        packages = g_list_remove(packages, npii);
      }
    }
  }
  
  g_list_free(l);
  return packages;
}

GList *
negotiate_fix_depends(GList *packages) {
  negotiate_package_info *npi, *npii;
  package_registration *rdep;
  GList *l, *item, *dep;
 
  l = g_list_copy(packages);
  
  for (item = l; item; item = item->next) {
    npi = (negotiate_package_info *)(item->data);
    
    for (dep = npi->reg->depends; dep; dep = dep->next) {
      rdep = (package_registration *)(dep->data);
      if (!(npii = negotiate_find_package(packages, rdep))) {
        debug_msg(2, "negotiate_fix_depends: package %s depends on %s, but %s is not supported", 
                  npi->reg->name, rdep->name, rdep->name);
        packages = g_list_remove(packages, npi);
        break;
      } else { 
        packages = g_list_remove(packages, npi);
        packages = g_list_append(packages, npi);
      }
    }
  }
  
  g_list_free(l);
  return packages; 
}

void
mcp_negotiate_simple(mcp_package * p, gchar * suffix, GList * fields) {
  gchar *pname;
  package_registration *reg;
  mcp_package *m;
  GList *packs;
  double version;
  gchar minVersion[10], maxVersion[10];
  gchar *minV, *maxV;
  negotiate_info *ni = (negotiate_info *)(p->user_data);
  negotiate_package_info *npi;
  
  if (strcasecmp(suffix, "can") == 0) {
    // Fields has package, min-version, max-version
    pname = mcp_find_value(fields, "package");
    reg = mcp_find_package(pname);

    if (reg) {
      minV = mcp_find_value(fields, "min-version");
      fix_decimal_point(minV, strlen(minV));
      maxV = mcp_find_value(fields, "max-version");
      fix_decimal_point(maxV, strlen(maxV));
      
      version =
        mcp_get_version(reg->minVersion, reg->maxVersion,
                        strtod(minV, NULL),
                        strtod(maxV, NULL));

      if (version > 0.0) {
        debug_msg(2, "MCP: %s, package is supported", pname);
        mcpconsole_add(p->session->owner, MCPC_MSG, "package is supported");
        
        npi = g_new(negotiate_package_info, 1);
        npi->reg = reg;
        npi->version = version;
        
        ni->packages = g_list_append(ni->packages, npi);
      } else {
        debug_msg(2, "MCP: %s, package supported but version not supported!",
                  pname);
        mcpconsole_add(p->session->owner, MCPC_MSG, "package version not supported, package is NOT accepted");
      }
    } else {
      debug_msg(2, "MCP: %s, package is not supported!", pname);
      mcpconsole_add(p->session->owner, MCPC_MSG, "package not supported");
    }
  } else if (strcasecmp(suffix, "end") == 0) {
    for (packs = mcp_get_registry(); packs; packs = packs->next) {
      reg = (package_registration *) (packs->data);

      snprintf(minVersion, 9, "%'.1f", reg->minVersion);
      fix_decimal_point_rev(minVersion, strlen(minVersion));
      
      snprintf(maxVersion, 9, "%'.1f", reg->maxVersion);
      fix_decimal_point_rev(maxVersion, strlen(maxVersion));

      mcp_send_simple(p->session, "mcp-negotiate-can", "package", reg->name,
                      "min-version", minVersion, "max-version", maxVersion,
                      NULL);
    }

    mcp_send_simple(p->session, "mcp-negotiate-end", NULL);
    
    ni->packages = negotiate_fix_overrides(ni->packages);
    ni->packages = negotiate_fix_depends(ni->packages);
    
    for (packs = ni->packages; packs; packs = packs->next) {
      npi = (negotiate_package_info *)(packs->data);
      
      if ((m = mcp_create_package_for(p->session, npi->reg))) {
        m->version = npi->version;

        if (npi->reg->initFunc) {
          npi->reg->initFunc(m);
        }
      }
      
      g_free(npi);
    }
    
    g_list_free(ni->packages);
    ni->packages = NULL;
  }
}
