#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "mcp.h"
#include "debug.h"
#include "packages.h"
#include "if_mcpconsole.h"

typedef struct _playerdb_signal_info {
  playerdb_signal_type signal_type;
  playerdb_signal_func callback;
  gchar *prop;
} playerdb_signal_info;

typedef struct _playerdb_info {
  gint self;
  GList *db;  
  GList *last_init_keys;
  GList *signals;
  mcp_package *package;
} playerdb_info;

void 
mcp_playerdb_init(mcp_package *p) {
  playerdb_info *pi = g_new(playerdb_info, 1);
  
  pi->self = -1;
  pi->db = NULL;
  pi->last_init_keys = NULL;
  pi->signals = NULL;
  pi->package = p;
  p->user_data = pi;
}

void
playerdb_destroy_hash_thingie(gpointer user_data) {
  g_free(user_data);
}

gboolean
mcp_playerdb_signal_connect(mcp_session *ses, playerdb_signal_type signal_type,
                            gchar *key, playerdb_signal_func callback) {
  playerdb_info *pi;
  playerdb_signal_info *signal;
  mcp_package *package;

  if (signal_type == SIG_SET && !key) {
    return FALSE;
  }
  
  package = 
    mcp_session_find_package(ses, "dns-nl-icecrew-playerdb");

  if (package) {
    pi = (playerdb_info *)(package->user_data);
    signal = g_new(playerdb_signal_info, 1);
     
    signal->signal_type = signal_type;
    signal->callback = callback;
      
    if (signal_type == SIG_SET) {
      signal->prop = g_strdup(key);
    } else {
      signal->prop = NULL;
    }
      
    pi->signals = g_list_append(pi->signals, signal);
    
    return TRUE;
  } else {
    return FALSE;
  }
}	

gboolean 
mcp_playerdb_signal_disconnect(mcp_session *ses, playerdb_signal_func callback) {
  GList *s, *signals;
  playerdb_signal_info *psi;
  playerdb_info *pi;
  
  mcp_package *package = 
    mcp_session_find_package(ses, "dns-nl-icecrew-playerdb");

  if (package) {
    pi = (playerdb_info *)(package->user_data);
    signals = g_list_copy(pi->signals);
    
    for (s = signals; s; s = s->next) {
      psi = (playerdb_signal_info *)(s->data);
      
      if (psi->callback == callback) {
        pi->signals = g_list_remove(pi->signals, psi);
        
        if (psi->prop) {
          g_free(psi->prop);
        }
        
        g_free(psi);
      }
    }
    
    g_list_free(signals);
    
    return TRUE;
  } else {
    return FALSE;
  }    
}

void
playerdb_call_signal(playerdb_signal_type signal_type, playerdb_info *pi, 
                     playerdb_player_info *ppi, gchar *key, gchar *value) {
  GList *l;
  playerdb_signal_info *psi;
  
  
  for (l = pi->signals; l; l = l ->next) {
    psi = (playerdb_signal_info *)(l->data); 
    
     if (psi->signal_type == signal_type) {
       if (signal_type == SIG_ADD || signal_type == SIG_DEL) {
         psi->callback(pi->package, ppi, NULL, NULL);
       } else if (strcasecmp(key, psi->prop) == 0) {
         psi->callback(pi->package, ppi, key, value);
       }
     }
  }
}

playerdb_player_info *
playerdb_find(GList *db, gint id) {
  playerdb_player_info *ppi;
  
  for (; db; db = db->next) {
    ppi = (playerdb_player_info *)(db->data);
    
    if (ppi->id == id) {
      return ppi;
    }
  }
  
  return NULL;
}

void 
mcp_playerdb_fini(mcp_package *p) {
  playerdb_info *pi = (playerdb_info *)(p->user_data);
  playerdb_player_info *ppi;
  playerdb_signal_info *psi;
  GList *f;
  
  for (f = pi->db; f; f = f->next) {
    ppi = (playerdb_player_info *)(f->data);
    g_hash_table_destroy(ppi->values);
    g_free(ppi);
  }
  
  g_list_free(pi->db);
  
  for (f = pi->signals; f; f = f->next) {
    psi = (playerdb_signal_info *)(f->data);
    
    if (psi->signal_type == SIG_SET) {
      g_free(psi->prop);
    }
    
    g_free(psi);
  }

  g_list_free(pi->signals);
    
  if (pi->last_init_keys) {
    g_free(pi->last_init_keys);
  }
}

gchar *
playerdb_get_prop(playerdb_player_info *ppi, gchar *key) {
  return g_hash_table_lookup(ppi->values, key);
}

void
playerdb_find_players_with(mcp_session *s, gchar *key, gchar *value, 
                           playerdb_find_func cb, gpointer user_data) {
  mcp_package *p = mcp_session_find_package(s, "dns-nl-icecrew-playerdb");
  playerdb_info *pi;
  playerdb_player_info *ppi;
  gchar *v;
  GList *db;
  
  if (p == NULL) {
    return;
  }
  
  pi = (playerdb_info *)(p->user_data);
  db = pi->db;
  
  for (; db; db = db->next) {
    ppi = (playerdb_player_info *)(db->data);
    
    v = playerdb_get_prop(ppi, key);
    
    if (v != NULL && (value == NULL || strcmp(v, value) == 0)) {
      cb(ppi, user_data);
    }
  }  
  
}

void
playerdb_set_prop(playerdb_info *pi, playerdb_player_info *ppi, gchar *key,
                  gchar *value, gboolean add) {
  debug_msg(2, "playerdb_set_prop: key: %s, value: %s", key, value);
  g_hash_table_insert(ppi->values, g_strdup(key), g_strdup(value));
  
  if (!add) {
    playerdb_call_signal(SIG_SET, pi, ppi, key, value);
  }
}

void
playerdb_modify_prop(playerdb_info *pi, playerdb_player_info *ppi, gchar *key, 
                     gchar *value, gboolean add) {
  if (ppi) {
    playerdb_set_prop(pi, ppi, key, value, add);
  }
}

void
playerdb_modify_prop_list(playerdb_info *pi, playerdb_player_info *ppi, 
                          GList *fields, gboolean add) {
  keyvaluepair *data;
  
  if (ppi) {
    for (; fields; fields = fields->next) {
      data = (keyvaluepair *)(fields->data);
      playerdb_set_prop(pi, ppi, data->key, data->value, add);
    }
  }
}

void
playerdb_remove_player(playerdb_info *pi, gint id) {
  playerdb_player_info *ppi;

  if ((ppi = playerdb_find(pi->db, id))) {
    playerdb_call_signal(SIG_DEL, pi, ppi, NULL, NULL);
    
    pi->db = g_list_remove(pi->db, ppi);
    g_hash_table_destroy(ppi->values);
  }
}

playerdb_player_info *
playerdb_add_player(playerdb_info *pi, int id) {
  playerdb_player_info *ppi = g_new(playerdb_player_info, 1);
  ppi->values = g_hash_table_new_full(g_str_hash, g_str_equal, 
                                      playerdb_destroy_hash_thingie, 
                                      playerdb_destroy_hash_thingie);
  ppi->id = id;
  pi->db = g_list_append(pi->db, ppi);
  
  debug_msg(2, "playerdb_add_player: adding id %d", id);
  
  return ppi;
}

void 
mcp_playerdb_simple(mcp_package * p, gchar * suffix, GList * fields) {
  gchar *value;
  playerdb_info *pi = (playerdb_info *)(p->user_data);
  playerdb_player_info *ppi;
  gint id, add;
      
  if (strcasecmp(suffix, "self") == 0) {
    if ((value = mcp_find_value(fields, "id"))) {
      pi->self = atoi(value);
    }
  } else if ((add = strcasecmp(suffix, "add")) == 0 || 
              strcasecmp(suffix, "set") == 0) {
    if ((value = mcp_find_value(fields, "id"))) {
      id = atoi(value);
      ppi = playerdb_find(pi->db, id);
      
      if (add == 0 && !ppi) {
        ppi = playerdb_add_player(pi, id);
        playerdb_modify_prop_list(pi, ppi, fields, TRUE);
        
        playerdb_call_signal(SIG_ADD, pi, ppi, NULL, NULL);
      } else if (add != 0 && ppi) {
        playerdb_modify_prop_list(pi, ppi, fields, FALSE);
      }
    }
  } else if (strcasecmp(suffix, "del") == 0) {
    if ((value = mcp_find_value(fields, "id"))) {
      id = atoi(value);
      playerdb_remove_player(pi, id);
    }
  } else if (strcasecmp(suffix, "init") == 0) {
    if ((value = mcp_find_value(fields, "keys"))) {
      if (pi->last_init_keys) {
        g_free(pi->last_init_keys);
      }
      pi->last_init_keys = mcp_parse_list(value);
    }
  }
}

gboolean 
mcp_playerdb_multi(mcp_package * p, gchar * data_tag, gchar * key, 
                   gchar * value, GList * allValues) {
  int i;
  gchar *id, *list;
  GList *l;
  playerdb_info *pi = (playerdb_info *)(p->user_data);
  playerdb_player_info *ppi;
    
  if (key != NULL) {
    /* We should have a value something like <NUM> <LIST> */
    i = 0;
    
    while (value[i] != '\0' && value[i] != ' ') {
      i++;
    }
    
    if (value[i] == '\0') {
      return TRUE;
    }
    
    id = g_strndup(value, i);
    value = mcp_jump_space(value + i);
    
    if (value[0] == '"') {
      value++;
      i = 0;
      
      while (value[i] != '\0' && value[i] != '"') {
        if (value[i] == '\\' && value[i] != '\0') {
          i++;
        }
        
        i++;
      }
      
      list = g_strndup(value, i);            
      l = mcp_parse_list(list);
            
      if (g_list_length(l) == g_list_length(pi->last_init_keys)) {
        i = atoi(id);
        
        if ((ppi = playerdb_find(pi->db, i))) {
          playerdb_remove_player(pi, i);
        }
        
        ppi = playerdb_add_player(pi, i);
        
        for (i = 0; i < g_list_length(l); i++) {
          playerdb_modify_prop(pi, ppi, 
                               g_list_nth_data(pi->last_init_keys, i),
                               g_list_nth_data(l, i), TRUE);
        }
        
        playerdb_call_signal(SIG_ADD, pi, ppi, NULL, NULL);
      }
      
      g_free(list);
      mcp_free_list(l);
    }
    
    g_free(id);
  } else {
    mcp_free_list(pi->last_init_keys);
    pi->last_init_keys = NULL;
  }
  
  return TRUE;
}

