#include <gtk/gtk.h>
#include <glib.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#include "debug.h"
#include "world.h"
#include "net.h"

/* These values taken from RFC 854 and RFC 857. */
#define TN_WILL 251
#define TN_WONT 252
#define TN_DO   253
#define TN_DONT 254
#define TN_IAC  255             /* Interpret As Command */
#define TN_SB   250             /* start of subnegotiation */
#define TN_SE   240             /* end of subnegotiaton */

gboolean net_input_recv(GIOChannel * source, GIOCondition condition,
                        gpointer data);

void net_connect_next(world *wld);

void
net_clean_disconnection(world * wld) {
  GError *err = NULL;

  if (!wld->net.channel) {
    debug_msg(1, "net_clean_disconnection: NOT clean for (%s), %d",
              options_get_str(wld->settings, "name"), wld->net.socket);
    return;
  }

  debug_msg(1, "net_clean_disconnection: clean disconnect for (%s), %d",
            options_get_str(wld->settings, "name"), wld->net.socket);

  // Pff, stupid, but clean the lot
  if (wld->net.input_buffer) {
    g_free(wld->net.input_buffer);
    wld->net.input_buffer = NULL;
  }
  // Shutdown the channel
  g_io_channel_shutdown(wld->net.channel, TRUE, &err);

  if (err) {
    debug_msg(1, "net_client_disconnection: error on channel shutdown: %s",
              err->message);
    g_error_free(err);
  }

  g_io_channel_unref(wld->net.channel);

  wld->net.channel = NULL;
  wld->net.socket = -1;

  wld->net.tn_last = 0;
  wld->net.tn_subneg = 0;

  world_writeln(wld, _("# Disconnected"));
  wld->net.state = NET_DISCONNECTED;
  
  world_disconnected(wld);
}

void
net_dirty_disconnection(world * wld, int err) {
  gchar *msg;
  
  // Pff, stupid, we print a message and pass it on to clean_disconnection
  debug_msg(1, "net_dirty_disconnection: dirty disconnect for (%s), %d",
            options_get_str(wld->settings, "name"), wld->net.socket);
            
  msg = g_strdup_printf(_("# Connection lost... (%s)"), strerror(err));
  world_writeln(wld, msg);
  g_free(msg);
  net_clean_disconnection(wld);
}

void
net_line_handle_telnet(world * wld, unsigned char *buf, int *len) {
  int i, j;
  unsigned char c;

  j = 0;
  for (i = 0; i < *len; i++) {
    c = buf[i];

    if (wld->net.tn_last) {
      switch (wld->net.tn_last) {
      case TN_WILL:
      case TN_WONT:
      case TN_DO:
      case TN_DONT:
        wld->net.tn_last = 0;
        break;
      case TN_IAC:
        switch (c) {
        case TN_WILL:
        case TN_WONT:
        case TN_DO:
        case TN_DONT:
          wld->net.tn_last = c;
          break;
        case TN_SB:
          wld->net.tn_subneg = 1;
          wld->net.tn_last = 0;
          break;
        case TN_SE:
          wld->net.tn_subneg = 0;
          wld->net.tn_last = 0;
          break;
        case TN_IAC:
          if (!wld->net.tn_subneg) {
            buf[j] = c;
            j++;
          }
          wld->net.tn_last = 0;
          break;
        default:
          wld->net.tn_last = 0;
        }
      }
    } else if (c == TN_IAC) {
      wld->net.tn_last = TN_IAC;
    } else if (wld->net.tn_subneg) {
      continue;
    } else {
      buf[j] = c;
      j++;
    }
  }

  *len = j;                     //Since j-- is the last written char
}

gboolean
net_input_recv(GIOChannel * source, GIOCondition condition, gpointer data) {
  world *wld = (world *) data;
  unsigned char rbuf[1024];
  unsigned char lbuf[1024];     // Will contain a line, used for breaking up rbuf in lines
  unsigned char *inp_buf;
  int len, i, j;

  if (condition == G_IO_HUP) {
    wld->net.state = NET_DISCONNECTING;
    net_clean_disconnection(wld);
    return FALSE;
  }

  if (wld->net.state != NET_CONNECTED) {
    debug_msg(1, "net_input_recv: world %s isn't even connected!",
              options_get_str(wld->settings, "name"));
    return FALSE;
  }
  // Break the received line by newline (skip \r)
  len = recv(wld->net.socket, lbuf, sizeof(lbuf) - 2, 0);

  debug_msg(1, "net_input_recv: received %d bytes", len);

  if (len < 1) {
    // Disconnected, either clean or dirty (shouldn't this be caught by G_IO_HUP?
    if (len < 0) {
      if (errno == EAGAIN || errno == EWOULDBLOCK) {
        return TRUE;
      }
      wld->net.state = NET_DISCONNECTING;
      net_dirty_disconnection(wld, errno);
    } else {
      wld->net.state = NET_DISCONNECTING;
      net_clean_disconnection(wld);
    }

    return FALSE;
  } else {
    // Its fine, we have a line :)
    net_line_handle_telnet(wld, lbuf, &len);

    j = 0;
    for (i = 0; i < len; i++) {
      if (lbuf[i] != '\r') {    // Skip the f*cking \r
        if (lbuf[i] == '\n') {  // End of line, finally!
          rbuf[j] = '\0';
          world_process_line(wld, rbuf);  //world_process_line should take care of the inputbuffer, also free IT!
          j = 0;
          rbuf[j] = '\0';
        } else {
          rbuf[j] = lbuf[i];
          j++;
        }
      }
    }

    // If j > 0 then we have some pending data! Store it!
    if (j > 0) {
      // Oooh, rbuf[j] != \0 but should be!
      rbuf[j] = '\0';
      if (wld->net.input_buffer) {  // Problem.. apparently we still have a buffer...
        inp_buf = g_strconcat(wld->net.input_buffer, rbuf, NULL);
        g_free(wld->net.input_buffer);
        wld->net.input_buffer = inp_buf;
      } else {
        wld->net.input_buffer = g_strdup(rbuf);
      }
    }
  }

  return TRUE;
}

void
net_connect_succeed(world *wld) {

  freeaddrinfo(wld->net.addr);
  wld->net.addr = NULL;
  wld->net.current = NULL;
  
  wld->net.source =
    g_io_add_watch(wld->net.channel, G_IO_IN | G_IO_HUP, net_input_recv, wld);

  wld->net.state = NET_CONNECTED;
  wld->manualDisconnect = FALSE;
  
  world_connected(wld);
}

void
net_connect_failed(world *wld, gchar *err) {
  gchar *message = g_strdup_printf(_("# Connect failed: %s"), err);
  
  world_writeln(wld, message);
  g_free(message);

  if (wld->net.channel) {
    g_io_channel_shutdown(wld->net.channel, TRUE, NULL);
    g_io_channel_unref(wld->net.channel);
    wld->net.channel = NULL;
  }
    
  if (wld->net.addr && wld->net.current->ai_next) {
    wld->net.current = wld->net.current->ai_next;
    net_connect_next(wld);
  } else {
    wld->net.socket = -1;
    
    if (wld->net.addr) {
      freeaddrinfo(wld->net.addr);
      wld->net.addr = NULL;
      wld->net.current = NULL;    
    }

    wld->net.state = NET_DISCONNECTED;
  }
}

gboolean
net_connect_check(GIOChannel * source, GIOCondition condition, gpointer data) {
  int option = 0, optionsize = sizeof(option);
  world *wld = (world *)data;
  
  if (getsockopt(wld->net.socket, SOL_SOCKET, SO_ERROR, &option, &optionsize) 
      < 0) {
    net_connect_failed(wld, strerror(errno));
  } else if (option) {
    net_connect_failed(wld, strerror(option));
  } else {
    net_connect_succeed(wld);
  }
  
  return FALSE;
}

void
net_connect_next(world *wld) {
  char host[NI_MAXHOST], port[NI_MAXSERV];
  unsigned char *message;
  int ret, result;
  struct addrinfo *tmp = wld->net.current;
  
  if (tmp == NULL) {
    return;
  } else {
       
    if ((ret =
         getnameinfo(tmp->ai_addr, tmp->ai_addrlen, host, NI_MAXHOST, port,
                     NI_MAXSERV, NI_NUMERICHOST | NI_NUMERICSERV)) != 0) {
      debug_msg(1, "net_connect_next: getnameinfo error: %s",
                gai_strerror(ret));
      net_connect_failed(wld, (gchar *)gai_strerror(ret));
      return;          
    }
    
    message = g_strdup_printf(_("# Trying %s port %s"), host, port);
    debug_msg(1, "net_connect_next: opening connection to %s port %s", host,
              port);
    world_writeln(wld, message);
    g_free(message);
    
    wld->net.socket = socket(tmp->ai_family, tmp->ai_socktype, tmp->ai_protocol);   
    
    if (wld->net.socket < 0) {
      net_connect_failed(wld, strerror(errno));
    } else {
      fcntl(wld->net.socket, F_SETFL, fcntl(wld->net.socket, F_GETFL) | O_NONBLOCK);
      
      if ((result = connect(wld->net.socket, tmp->ai_addr, 
          wld->net.addr->ai_addrlen)) == -1 && errno != EINPROGRESS) {          
        net_connect_failed(wld, strerror(errno));
      } else {
        // Its not done yet, add it to the glib loop
        wld->net.channel = g_io_channel_unix_new(wld->net.socket);
        g_io_channel_set_close_on_unref(wld->net.channel, TRUE);
        g_io_add_watch(wld->net.channel, G_IO_IN | G_IO_HUP, 
                       net_connect_check, wld);
      }        
    }
  }
}

void
net_open_connection(world * wld, gchar *h, gchar *p) {
  struct addrinfo hint;
  int ret;

  char host[NI_MAXHOST], port[NI_MAXSERV];

  snprintf(host, NI_MAXHOST - 1, "%s", h);
  snprintf(port, NI_MAXSERV - 1, "%s", p);

  memset(&hint, 0, sizeof(hint));
  hint.ai_flags = 0;
  hint.ai_family = AF_UNSPEC;
  hint.ai_socktype = SOCK_STREAM;
  hint.ai_protocol = IPPROTO_TCP;

  if ((ret = getaddrinfo(host, port, &hint, &(wld->net.addr))) != 0) {
    wld->net.addr = NULL;
    debug_msg(1, "net_open_connection: getaddrinfo failed: %s", gai_strerror(ret));
    net_connect_failed(wld, (gchar *)gai_strerror(ret));
    return;
  }

  if (wld->net.addr) {
    wld->net.current = wld->net.addr;
    net_connect_next(wld);
  } else {
    net_connect_failed(wld, _("No addresses available"));
  }
}

void
net_send_line(world * wld, unsigned char *line) {
  unsigned char *sendLine;
  int result;
  fd_set fdConnectSet;

  if (wld->net.state == NET_CONNECTED) {
    debug_msg(1, "net_send_line: %s", line);

    if (strncmp(line, "#$#", 3) != 0) {
      g_free(wld->prevLine);
      wld->prevLine = g_strdup(line);
    }

    sendLine = g_strconcat(line, "\r\n", NULL);

    if ((result = send(wld->net.socket, sendLine, strlen(sendLine), 0)) == -1
        && (errno == EAGAIN || errno == EWOULDBLOCK)) {
      FD_ZERO(&fdConnectSet);
      FD_SET(wld->net.socket, &fdConnectSet);

      // Wait for sending to be done
      select(wld->net.socket + 1, NULL, &fdConnectSet, NULL, NULL);
    } else if (result == -1) {
      debug_msg(1, "net_send_line: error on sending line: %s", strerror(errno));
      net_dirty_disconnection(wld, errno);
    }

    g_free(sendLine);
  } else {
    world_writeln(wld, _("# Not connected"));
    debug_msg(1, "net_send_line: world (%s) not connected!",
              options_get_str(wld->settings, "name"));
  }
}

void 
net_connect_to(world *wld, gchar *host, gchar *port) {
  unsigned char message[1024];

  if (wld->net.state != NET_DISCONNECTED) {
    return;
  }

  if (wld->tryReconnect != -1) {
    g_source_remove(wld->tryReconnect);
    wld->tryReconnect = -1;
  }

  wld->net.state = NET_CONNECTING;
  snprintf(message, 1024 - 1, _("# Connecting to %s port %s"), host, port);
           
  world_writeln(wld, message);
  net_open_connection(wld, host, port);
}

void
net_do_connect(world * wld) {
  net_connect_to(wld, options_get_str(wld->settings, "host"), 
                 options_get_str(wld->settings, "port"));
}

void
net_do_disconnect(world * wld) {
  if (wld->net.state == NET_CONNECTED) {
    wld->manualDisconnect = TRUE;
    wld->net.state = NET_DISCONNECTING;
    world_writeln(wld, _("# Disconnecting..."));

    // Remove the watch
    g_source_remove(wld->net.source);

    net_clean_disconnection(wld);
  }
}
