#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include <gtk/gtk.h>
#include <regex.h>
#include <libgnome/gnome-url.h>
#include <libgnomevfs/gnome-vfs.h>

#include "support.h"
#include "debug.h"
#include "if_main.h"

#define URL_REGEXP "(((mailto|news|telnet|nttp|file|http|sftp|ftp|https|dav|callto)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.@:]+[^]''\\.}>\\) ,\\/\\\"\\!]+(:[0-9]*)?(/[-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:@&=\\?/~\\#\\%]*[^]'\\.}>\\) ,\\\"\\!])?"
static regex_t url_regexp;

gchar *
fix_decimal_point(gchar *line, int len) {
  int i;
  struct lconv *l = localeconv();

  for (i = 0; i < len; i++) {
    if (line[i] == '.') {
      line[i] = l->decimal_point[0];
    }
  }
  
  return line;
}

gchar *
fix_decimal_point_rev(gchar *line, int len) {
  int i;
  struct lconv *l = localeconv();

  for (i = 0; i < len; i++) {
    if (line[i] == l->decimal_point[0]) {
      line[i] = '.';
    }
  }

  return line;
}

gchar *
ansi_strip(gchar * s) {
  int i, j = 0;

  for (i = 0; s[i] != '\0'; i++) {
    // Escape sequence, advance to character after 'm'
    if (s[i] == '\x1B') {
      while (s[i] != '\0' && s[i] != 'm') {
        i++;
      }
    } else if (s[i] != '\x07') {
      s[j] = s[i];
      j++;
    }
  }

  s[j] = '\0';
  return s;
}

int 
garray_length(gchar **s) {
  int i = 0;
  
  while (s[i] != NULL) {
    i++;
  }
  
  return i;
}

void g_list_free_simple(GList *s) {
  GList *tmp;
  
  for (tmp = s; tmp; tmp = tmp->next) {
    g_free(tmp->data);  
  }
  
  g_list_free(s);
}

gchar *
g_list_find_simple(GList *s, gchar *f) {
  GList *tmp;
  
  for (tmp = s; tmp; tmp = tmp->next) {
    if (strcmp(tmp->data, f) == 0) {
      return tmp->data;
    } 
  }
  
  return NULL;
}

void
gnoemoe_dialog(gchar * message, GtkMessageType messagebox_type,
               GtkWindow * parent) {
  GtkWidget *dlg;

  if (parent == NULL) {
    parent = GTK_WINDOW(if_main_get_widget("wndMain"));
  }

  dlg =
    gtk_message_dialog_new(parent,
                           GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                           messagebox_type, GTK_BUTTONS_OK, message, NULL);
  gtk_dialog_run(GTK_DIALOG(dlg));
  gtk_widget_destroy(dlg);
}

void
gnoemoe_error_dialog(gchar * message, GtkWindow * parent) {
  gnoemoe_dialog(message, GTK_MESSAGE_ERROR, parent);
}

void
gnoemoe_warning_dialog(gchar * message, GtkWindow * parent) {
  gnoemoe_dialog(message, GTK_MESSAGE_WARNING, parent);
}

void
gnoemoe_info_dialog(gchar * message, GtkWindow * parent) {
  gnoemoe_dialog(message, GTK_MESSAGE_INFO, parent);
}

void
gnoemoe_question_dialog(gchar * message, GtkWindow * parent) {
  gnoemoe_dialog(message, GTK_MESSAGE_QUESTION, parent);
}

void
doEvents() {
  while (gtk_events_pending()) {
    gtk_main_iteration();
  }
}

gchar *
str_escape(gchar * line) {
  gchar *newLine;
  int i, j = 0;

  if (strlen(line) == 0) {
    return g_strdup("");
  }

  if (strstr(line, "\"") || strstr(line, "\\")) {
    newLine = g_new(gchar, (strlen(line) * 2) + 1);

    for (i = 0; i < strlen(line); i++) {
      if (line[i] == '"' || line[i] == '\\') {
        newLine[j] = '\\';
        j++;
      }
      newLine[j] = line[i];
      j++;
    }

    newLine[j] = '\0';
    return newLine;
  } else {
    return g_strdup(line);
  }
}

gint
gnoemoe_url_regex_match(const gchar *msg, int len, GArray *start, GArray *end) {
  static gboolean inited = FALSE;
  regmatch_t matches[1];
  gint ret = 0, num_matches = 0, offset = 0;
  gchar *tmp;
  
  if (!inited) {
    memset (&url_regexp, 0, sizeof (regex_t));
    regcomp (&url_regexp, URL_REGEXP, REG_EXTENDED);
    inited = TRUE;
  }

  tmp = g_strndup(msg, len);

  while (!ret) {
    ret = regexec(&url_regexp, (char *)(tmp + offset), 1, matches, 0);

    if (ret == 0) {
      if (matches[0].rm_so > matches[0].rm_eo) {
        break;
      }
      
      gint s;
      num_matches++;

      s = matches[0].rm_so + offset;
      offset = matches[0].rm_eo + offset;
		
      g_array_append_val(start, s);
      g_array_append_val(end, offset);
    }
  }
	
  g_free(tmp);

  return num_matches;
}

void
gnoemoe_open_url (const gchar *url) {
	if (!url || strlen (url) == 0) {
		return;
	}

	/* gnome_url_show doesn't work when there's no protocol, so we might
	 * need to add one.
	 */
	if (strstr (url, "://") == NULL) {
		gchar *tmp;
		
		tmp = g_strconcat ("http://", url, NULL);
		gnome_url_show(tmp, NULL);
		g_free (tmp);
		return;
	}
	
	gnome_url_show (url, NULL);
}

void
gnoemoe_fetch_handle_free(GnoeMoeFetchHandle *g) {
  GList *tmp;
  
  g_free(g->cur_file_name);
  
  for (tmp = g->source_uri; tmp; tmp = tmp->next) {
    gnome_vfs_uri_unref((GnomeVFSURI*)(tmp->data));
  }  
	g_list_free(g->source_uri);
	

  for (tmp = g->dest_uri; tmp; tmp = tmp->next) {
    gnome_vfs_uri_unref((GnomeVFSURI*)(tmp->data));
  }  
	g_list_free(g->dest_uri);

	g_free(g);
}

GnoeMoeFetchHandle *
gnoemoe_fetch_handle_create(GFunc cb, gpointer user_data) {
  GnoeMoeFetchHandle *g = g_new0(GnoeMoeFetchHandle, 1);
  
  g->cb = cb;
  g->user_data = user_data;
	g->prev_status = GNOME_VFS_XFER_PROGRESS_STATUS_OK;
	g->cur_phase = -1;
	g->prev_phase = -1;
	g->cur_file = -1;
	g->prev_file = -1;
	g->source_uri = NULL;
	g->dest_uri = NULL;
	g->cur_file_name = NULL;
  g->files_total = 0;
	g->done = FALSE;
	g->aborted = FALSE;
	
	return g;
}

gint
gnoemoe_fetch_progress(GnomeVFSAsyncHandle *handle, 
                       GnomeVFSXferProgressInfo *info,
                       GnoeMoeFetchHandle *g) {
  gchar *name;
  const gchar *err;

	g->cur_phase = info->phase;
	g->cur_file = info->file_index;
	g->files_total = info->files_total;
	g->bytes_total = info->bytes_total;
	g->file_size = info->file_size;
	g->bytes_copied = info->bytes_copied;
	g->total_bytes_copied = info->total_bytes_copied;
		
  if (g->aborted) {
    g->cb(g, g->user_data);
    gnoemoe_fetch_handle_free(g);
    return FALSE;
  }

  if (info->target_name != NULL) {
    if (g->cur_file_name && strcmp(g->cur_file_name, info->target_name) != 0) {
      g->cur_phase = GNOME_VFS_XFER_PHASE_FILECOMPLETED;
      g->cb(g, g->user_data);
      g->cur_phase = info->phase;    
    
      g_free(g->cur_file_name);
      g->cur_file_name = NULL;
    }

    if (!g->cur_file_name) {
      g->cur_file_name = g_strdup(info->target_name);
    }
  }
	   
  if (info->status == GNOME_VFS_XFER_PROGRESS_STATUS_OVERWRITE) {
    name = gnome_vfs_get_local_path_from_uri(info->target_name);
    debug_msg(1, "gnoemoe_fetch_progress: asking for overwriting %s: yes", name);
    
    g->prev_status = GNOME_VFS_XFER_PROGRESS_STATUS_OVERWRITE;
    return GNOME_VFS_XFER_OVERWRITE_ACTION_REPLACE;
  } else if (info->status == GNOME_VFS_XFER_PROGRESS_STATUS_VFSERROR) {
    name = gnome_vfs_get_local_path_from_uri(info->target_name);
    err = gnome_vfs_result_to_string(info->vfs_status); 
     
    gdk_threads_enter();
    debug_msg(1, "gnoemoe_fetch_progress: error for %s: %s", name, err);
    g->cb(g, g->user_data);
    gdk_threads_leave();
    
    g_free(g->cur_file_name);
    g->cur_file_name = NULL;
     
    g->prev_status = GNOME_VFS_XFER_PROGRESS_STATUS_VFSERROR;
    g_free(name);
    
    return GNOME_VFS_XFER_ERROR_ACTION_SKIP;
  }
  
  if (info->phase == GNOME_VFS_XFER_PHASE_COMPLETED) {
    if (g->cur_file_name != NULL) {
      g->cur_phase = GNOME_VFS_XFER_PHASE_FILECOMPLETED;
      g->cb(g, g->user_data);
      g->cur_phase = info->phase;
    }
    
    g->done = TRUE;
    g->cb(g, g->user_data);
    gnoemoe_fetch_handle_free(g);
    return TRUE;  
  }
  
  g->prev_status = info->status;
  
  return TRUE;
}

gint 
gnoemoe_fetch_interact(GnomeVFSXferProgressInfo *info, gpointer user_data) {
  return 1;
}

GnoeMoeFetchHandle *
gnoemoe_fetch(const GList *source, const GList *dest, 
              GFunc cb, gpointer user_data) {
  GnoeMoeFetchHandle *g = gnoemoe_fetch_handle_create(cb, user_data);
  gchar *uri;
  
  for (; source; source = source->next) {
    uri = (gchar *)(source->data);
    g->source_uri = g_list_append(g->source_uri, gnome_vfs_uri_new(uri));
  }
  
  for (; dest; dest = dest->next) {
    uri = (gchar *)(dest->data);
    g->dest_uri = g_list_append(g->dest_uri, gnome_vfs_uri_new(uri));
  }
  
  gnome_vfs_async_xfer(&(g->handle), g->source_uri, g->dest_uri, 
                       GNOME_VFS_XFER_DEFAULT,
                       GNOME_VFS_XFER_ERROR_MODE_QUERY,
                       GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
                       GNOME_VFS_PRIORITY_DEFAULT,
                       (GnomeVFSAsyncXferProgressCallback)gnoemoe_fetch_progress,
                       g, gnoemoe_fetch_interact, g);
  return g;
}
