#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "ansi.h"
#include "debug.h"
#include "g_options.h"
#include "if_world.h"
#include "if_main.h"
#include "if_mcpconsole.h"
#include "main.h"
#include "mcp/mcp.h"
#include "net.h"
#include "my-string.h"
#include "support.h"
#include "gnoe_pixbuf.h"
#include "world.h"
#include "editor.h"

#ifdef HASRUBY
#  include "script.h"
#endif

void world_update_background(world *wld);

/** \ingroup world
  * \brief list containing all worlds
  *
  * list which contains all the worlds
  */
GList *worlds = NULL;

/** \ingroup world
  * \brief \b PRIVATE: create default settings
  *
  * \b PRIVATE: create the default settings for the world. Usually called when a world
  * is created.
  * @param wld the world to create the default settings for
  * @see world_create
  */
void
world_create_default_settings(world * wld) {
  const gchar *loc;
  g_get_charset(&loc);
  
  wld->settings = options_create();

  options_add(wld->settings, "name", "");
  options_add(wld->settings, "autoload", "0");
  options_add(wld->settings, "host", "");
  options_add(wld->settings, "port", "1111");
  options_add(wld->settings, "player_name", "");
  options_add(wld->settings, "reconnect", "0");
  options_add(wld->settings, "password", "");
  options_add(wld->settings, "text_buffer", "1000");
  options_add(wld->settings, "input_buffer", "500");
  options_add(wld->settings, "charset", (gchar *)loc);

  // Non configurable options
  options_add(wld->settings, "pane_position", "150");
}

/** \ingroup world
  * \brief \b PRIVATE: load the input history
  *
  * \b PRIVATE: load the input history for the world specified. The input history is 
  * located at #{wld->path}/history.  Usually called when a world
  * is created.
  * @param wld the world for which to load the input history
  * @see world_create
  */
void
world_load_input_history(world * wld) {
  FILE *f;
  gchar line[1024], *filename;

  filename = g_strconcat(wld->path, "/history", NULL);

  debug_msg(1, "world_load_input_history: loading history (%s)!", filename);

  if ((f = fopen(filename, "r")) != NULL) {
    while (fgets((char *) &line, 1024 - 1, f) != NULL) {
      line[strlen((char *) &line) - 1] = '\0';
      if (strlen(line) != 0) {  // Empty lines, we don't need to process those
        wld->entryBuffer.buffer =
          g_list_append(wld->entryBuffer.buffer, g_strdup(line));
      }
    }
    wld->entryBuffer.curPos = g_list_length(wld->entryBuffer.buffer);
    fclose(f);
  } else {
    debug_msg(1,
              "world_load_input_history: could not retrieve contents of file %s (%s)",
              filename, strerror(errno));
  }

  g_free(filename);
}

/** \ingroup world
  * \brief creates a new world
  *
  * create a new world struct from path. Load the settings (if path != NULL),
  * null all fields which need to be nulled and load the input history (if path
  * != NULL)
  * @param path the path to the worlds data directory structure, 
  * or NULL if this world is totally new
  * @return the newly created world. Must be destroyed by 
  * \link world::world_destroy world_destroy \endlink
  * @see world_create_default_settings
  * @see world_load_input_history
  * @see world_destroy
  */
world *
world_create(char *path) {
  world *result = (world *) malloc(sizeof(world));

  result->path = NULL;

  world_create_default_settings(result);

  if (path != NULL) {
    char *worldSettings = g_strconcat(path, "/settings", NULL);

    debug_msg(1, "world_create: creating new world for %s", path);
    result->path = g_strdup(path);
    debug_msg(1, "world_create: creating default world settings for %s", path);
    options_load(result->settings, worldSettings);    

    g_free(worldSettings);
  }

  result->net.channel = NULL;
  result->net.socket = -1;
  result->net.input_buffer = NULL;
  result->net.tn_last = 0;
  result->net.tn_subneg = 0;
  result->net.state = NET_DISCONNECTED;
  
  result->net.addr = NULL;
  result->net.current = NULL;

  result->loaded = 0;

  result->entryBuffer.buffer = NULL;
  result->entryBuffer.curPos = 0;
  result->entryBuffer.prevKey = 0;
  result->entryBuffer.prefix = NULL;
 
  if (path) {
    world_load_input_history(result);
  }

  result->editing = NULL;

  result->widgets.noteContainer = NULL;
  result->inactive = 0;

  result->prevLine = NULL;
  result->tryReconnect = -1;
  result->updateBg = -1;
  result->manualDisconnect = FALSE;

  result->reColor = 0;
  result->needScroll = FALSE;
 
  result->characterWidth = 0;
  result->characterHeight = 0;
   
  result->userlistScrollHandler = -1;
  return result;
}

/** \ingroup world
  * \brief copies a world
  *
  * create a new world based on a given world. The path and the settings are
  * copied
  * @param source the world to be copied
  * @return a new world which has the same settings and path as the source world
  */
world *
world_copy(world * source) {
  world *result = world_create(NULL);
  int i;

  result->path = g_strdup(source->path);
  // Copy the options now

  result->settings->filename = g_strdup(source->settings->filename);

  for (i = 0; i < source->settings->optionCount; i++) {
    options_set_str(result->settings, source->settings->values[i]->key,
                    source->settings->values[i]->value);
  }

  return result;
}

/** \ingroup world
  * \brief \b PRIVATE: saves input history to file
  *
  * \b PRIVATE: saves the current input history list to the 
  * file #{wld->path}/history
  * @param wld the world for which to save the input history
  */
void
world_save_input_history(world * wld) {
  FILE *f;
  gchar *filename;
  GList *elem;

  if (wld->path == NULL) {
    return;
  }

  filename = g_strconcat(wld->path, "/history", NULL);
  f = fopen(filename, "w");

  if (f) {
    debug_msg(1, "world_save_input_history: saving input history to %s",
              filename);
    for (elem = wld->entryBuffer.buffer; elem; elem = elem->next) {
      fprintf(f, "%s\n", (gchar *) (elem->data));
    }

    fclose(f);

    chmod(filename, 0660);
  } else {
    debug_msg(1,
              "world_save_input_history: couldn't open history file (%s) for saving: %s",
              filename, strerror(errno));
  }

  g_free(filename);
}

/** \ingroup world
  * \brief destroys a world
  *
  * destroys a world. This will save the world settings, save the input history
  * remove all files in the #{wld->path}/tmp directory and free any memory
  * allocated for the world
  * @param wld the world to destroy
  */
void
world_destroy(world * wld) {
  gchar *tmpDir;

  debug_msg(1, "world_destroy: destroying world %s",
            options_get_str(wld->settings, "name"));

  options_save(wld->settings);

  // Saving history
  world_save_input_history(wld);

  // Removing all tmp files
  tmpDir = g_strconcat(wld->path, "/tmp", NULL);
  main_directory_remove_all(tmpDir, FALSE);
  g_free(tmpDir);

  g_free(wld->path);
  options_destroy(wld->settings);

  g_list_free(wld->entryBuffer.buffer);
  g_free(wld->entryBuffer.prefix);
  g_free(wld->prevLine);
}

/** \ingroup world
  * \brief \b PRIVATE: adds text to the text view with \i tags applied
  *
  * \b PRIVATE: adds text to the text view with \i tags applied to the \i text
  * @param wld the world to add the text for
  * @param text the text to be added
  * @param len the length of the text
  * @param tags a list containing GtkTextTag * to apply to the text
  */
void
world_add_text(world * wld, char *text, int len, GList * tags) {
  GtkTextIter startIter, endIter;
  GtkTextBuffer *tb = GTK_TEXT_BUFFER(wld->widgets.bufVwWorld);
  gint start_offset;
  gint num_matches, i;
  GArray *start, *end;
  gint s = 0, e = 0, last = 0;
	
  gtk_text_buffer_get_end_iter(tb, &endIter);
  start_offset = gtk_text_iter_get_offset(&endIter);
  
  start = g_array_new(FALSE, FALSE, sizeof (gint));
  end = g_array_new(FALSE, FALSE, sizeof (gint));
  num_matches = gnoemoe_url_regex_match(text, len, start, end);
  
  if (num_matches == 0) {
    gtk_text_buffer_insert(tb, &endIter, text, len);
  } else {
    for (i = 0; i < num_matches; i++) {
      s = g_array_index (start, gint, i);
      e = g_array_index (end, gint, i);  

      if (s > last) {
        gtk_text_buffer_insert(tb, &endIter, text + last, s - last);
      }
			
      gtk_text_buffer_insert_with_tags_by_name(tb, &endIter, text + s, e - s,
			                                         "url", NULL);
      last = e;
    }
    
    if (e < len) {
      gtk_text_buffer_insert(tb, &endIter, text + e, len - e);
    }
  }
  
  g_array_free(start, TRUE);
  g_array_free(end, TRUE);
  
  if (tags) {
    gtk_text_buffer_get_iter_at_offset(tb, &startIter, start_offset);

    for (; tags; tags = tags->next) {
      gtk_text_buffer_apply_tag(tb, GTK_TEXT_TAG(tags->data), &startIter,
                                &endIter);
    }
  }

}

/** \ingroup world
  * \brief \b PRIVATE: removes tags from list with tags
  *
  * \b PRIVATE: remove all GtkTextTags from \i tags with the property \i prop
  * set to TRUE.
  * @param tags the list with GtkTextTags to remove tags from
  * @param prop the name of the property to check for 
  * @return the list with tags removed
  */
GList *
world_remove_tags_with_property(GList * tags, gchar * prop) {
  GList *t = g_list_copy(tags);
  GtkTextTag *tag;
  gboolean isSet;
  gchar *name = NULL;
  gboolean isForeground = (strcmp(prop, "foreground-set") == 0);
  
  for (; t; t = t->next) {
    tag = GTK_TEXT_TAG(t->data);
    g_object_get(G_OBJECT(t->data), prop, &isSet, NULL);
    
    if (isForeground) {
      g_object_get(G_OBJECT(t->data), "name", &name, NULL);
    }

    if (isSet || (isForeground && name[strlen(name) - 1] == 'h')) {
      tags = g_list_remove(tags, t->data);
    }
  }

  g_list_free(t);
  return tags;
}

/** \ingroup world
  * \brief \b PRIVATE: find name for ansi_code
  *
  * \b PRIVATE: find the name of the GtkTextTag corresponding with
  * the ansi code \i code
  * @param code the code to lookup the name for
  * @return the tag name of \i code
  */
gchar *
world_ansi_find_name(ansi_code code) {
  int i, len;

  len = (sizeof(ansi_colors) / sizeof(ansinamepair));

  for (i = 0; i < len; i++) {
    if (ansi_colors[i].code == code) {
      return (gchar *) ansi_colors[i].name;
    }
  }

  return NULL;
}

/** \ingroup world
  * \brief \b PRIVATE: add bold tags for colors
  *
  * \b PRIVATE: adds the equivalent bold tag for every color tag in the list
  * @param tags the list of GtkTextTags
  * @return the list of tags
  */
GList *
world_fix_tags_bold(GtkTextTagTable *tab, GList *tags) {
  GList *t;
  GtkTextTag *tag, *htag;
  gboolean isSet;
  gchar *name, *hname;
  gboolean added = FALSE;
  
  for (t = g_list_copy(tags); t; t = t->next) {
    tag = GTK_TEXT_TAG(t->data);
    
    g_object_get(G_OBJECT(tag), "foreground-set", &isSet, NULL);
    g_object_get(G_OBJECT(tag), "name", &name, NULL);
    
    if (isSet) {
      hname = g_strconcat(name, "_h", NULL);
      htag = gtk_text_tag_table_lookup(tab, hname);
      
      if (htag) {
        tags = g_list_append(tags, htag);
        added = TRUE;
      }
    }
  }
  
  /* If there is no high color added than this means we have a default color */
  if (!added) {
    tags = g_list_append(tags, gtk_text_tag_table_lookup(tab, 
                           world_ansi_find_name(A_FG_DEFAULT_H)));
  }
  
  g_list_free(t);
  return tags;
}

/** \ingroup world
  * \brief \b PRIVATE: remove tags for a certain ansi code
  *
  * \b PRIVATE: prepares \i tags for insertion of \i code by
  * removing all GtkTextTags from \i tags which are overriden by \i code
  * @param tags the list of GtkTextTags
  * @param code the ansi code
  * @return the list of tags
  */
GList *
world_remove_tags(GList * tags, ansi_code code) {
  switch (code) {
    case A_DEFAULT:
      g_list_free(tags);
      tags = NULL;
      break;
    case A_FG_DEFAULT: case A_FG_BLACK: case A_FG_RED: case A_FG_GREEN: 
    case A_FG_YELLOW: case A_FG_BLUE: case A_FG_PURPLE: case A_FG_CYAN:
    case A_FG_WHITE:
      tags = world_remove_tags_with_property(tags, "foreground-set");
      break;
    case A_BG_DEFAULT: case A_BG_BLACK: case A_BG_RED: case A_BG_GREEN:
    case A_BG_YELLOW: case A_BG_BLUE: case A_BG_PURPLE: case A_BG_CYAN:
    case A_BG_WHITE:
      tags = world_remove_tags_with_property(tags, "background-set");
      break;
    case A_BOLD: case A_FAINT: case A_BOLD_OFF:
      tags = world_remove_tags_with_property(tags, "weight-set");
      break;
    case A_UNDERLINE: case A_DOUBLE_UNDERLINE: case A_UNDERLINE_OFF:
      tags = world_remove_tags_with_property(tags, "underline-set");
      break;
    case A_ITALIC: case A_ITALIC_OFF:
      tags = world_remove_tags_with_property(tags, "style-set");
      break;
    case A_NOWRAP:
      tags = world_remove_tags_with_property(tags, "wrap-mode-set");
    default:
      break;
  }

  return tags;
}

/** \ingroup world
  * \brief \b PRIVATE: check/remove lines that exceed the buffer length
  *
  * \b PRIVATE: removes top lines from the text view of the world as long as
  * the number of lines exceeds the number of lines defined by the text_buffer
  * option in the world settings
  * @param wld the world to check the number of lines for
  */
void
world_check_buffer(world * wld) {
  GtkTextBuffer *buf = GTK_TEXT_BUFFER(wld->widgets.bufVwWorld);
  GtkTextIter start, end;
  int buffer = options_get_int(wld->settings, "text_buffer");

  while (gtk_text_buffer_get_line_count(buf) > buffer) {
    // Get start iter
    gtk_text_buffer_get_iter_at_line(buf, &start, 0);
    // Get end iter
    gtk_text_buffer_get_iter_at_line(buf, &end, 1);

    // Remove the line
    gtk_text_buffer_delete(buf, &start, &end);
  }
}

/** \ingroup world
  * \brief \b PRIVATE: recolor timeout callback
  *
  * \b PRIVATE: callback function for the recolor timeout. This function colors
  * the world tab label blue after a period of time when it was red colored (
  * new input)
  * @param wld the world to recolor the tab for
  * @result FALSE (remove the timeout source from the glib main loop)
  */
gboolean
on_world_recolor(world * wld) {
  GdkColor col;
  gchar *color;

  if (wld->reColor) {
    wld->reColor = 0;
    color = g_strdup_printf("#0000%02x", CALC_COLOR_RANGE(wld->inactive));
    gdk_color_parse(color, &col);
    g_free(color);

    gtk_widget_modify_fg(wld->widgets.mainTabInfo.tabLabel, GTK_STATE_ACTIVE,
                         &col);
  }

  return FALSE;
}

/** \ingroup world
  * \brief add a line to the text view output
  *
  * adds a line to the text view output and parses any ansi on the run 
  * @param wld the world to add the line to
  * @param line the line to write
  */
void
world_writeln(world * wld, unsigned char *line) {
  gchar *newLine, *newLineTerm, *ptr, *ansiStart, *ansiStop, *ansi, **ansis;
  GError *err = NULL;
  int written, read, seq, i;
  GList *tags = NULL;
  GtkTextTagTable *tab =
    gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(wld->widgets.bufVwWorld));
  GdkColor col;
  gchar *title, *color, *name, *nameh;
  gboolean bold = FALSE, active, current;
  
  if (!(wld->loaded)) {
    return;
  }

  world_scroll_end_prepare(wld);

  // Check for beeeeeeEEEEp!
  if (strchr(line, '\x07')) {
    gdk_beep();
    mystring_remove_char(line, '\x07'); // Remove all BEEEEP chars      
  }

  debug_msg(1, "world_writeln: %s", line);

  newLine =
    g_convert_with_fallback(line, strlen(line), "UTF-8",
                            options_get_str(wld->settings, "charset"),
                            NULL, &read, &written, &err);

  if (newLine) {
    newLineTerm = g_strconcat(newLine, "\n", NULL);
  } else {
    debug_msg(1, "world_writeln: conversion failed: %s (written %d, read %d)!",
              err->message, written, read);
    g_error_free(err);
    newLineTerm = g_strconcat(line, "\n", NULL);
  }

  ptr = newLineTerm;

  world_check_buffer(wld);

  while ((ansiStart = strchr(ptr, '\x1B'))) {
    if (ptr != ansiStart) {
      world_add_text(wld, ptr, ansiStart - ptr, tags);
    }
    // Find ansiStop
    if ((ansiStop = strchr(ansiStart, 'm'))) {
      ansiStart = ansiStart + 2;  // Advance to actual ansi sequence
      ansi = g_strndup(ansiStart, ansiStop - ansiStart);
      ansis = g_strsplit(ansi, ";", -1);
      g_free(ansi);

      for (i = 0; ansis[i]; i++) {
        if (strlen(ansis[i]) && mystring_to_int(ansis[i], &seq)) {
          // seq contains the new ansi code
          // First remove all tags from the tags list that are overridden by this tag
          tags = world_remove_tags(tags, seq);

          switch (seq) {
            case A_FG_DEFAULT: case A_FG_BLACK: case A_FG_RED: case A_FG_GREEN:
            case A_FG_YELLOW: case A_FG_BLUE: case A_FG_PURPLE: case A_FG_CYAN:
            case A_FG_WHITE:
              name = world_ansi_find_name(seq);
              nameh = g_strconcat(name, "_h", NULL);
              tags =
                g_list_append(tags,
                              gtk_text_tag_table_lookup(tab, name));
              if (bold) {
                // Add the bold color taggie
                tags = g_list_append(tags, 
                                     gtk_text_tag_table_lookup(tab, nameh));                                    
              }
            
              g_free(nameh);
              break;
            case A_BG_DEFAULT: case A_BG_BLACK: case A_BG_RED: case A_BG_GREEN:
            case A_BG_YELLOW: case A_BG_BLUE: case A_BG_PURPLE: case A_BG_CYAN:
            case A_BG_WHITE:
              tags =
                g_list_append(tags,
                              gtk_text_tag_table_lookup(tab,
                                                        world_ansi_find_name
                                                        (seq)));
              break;
            case A_DEFAULT:
              bold = FALSE;
              break;
            case A_BOLD:
              bold = TRUE;
              tags = g_list_append(tags, gtk_text_tag_table_lookup(tab, "bold"));
              tags = world_fix_tags_bold(tab, tags);
              break;
            case A_FAINT:
              tags =
                g_list_append(tags, gtk_text_tag_table_lookup(tab, "faint"));
              break;
            case A_BOLD_OFF:
              bold = FALSE;
              tags =
                g_list_append(tags, gtk_text_tag_table_lookup(tab, "bold-off"));
              break;
            case A_UNDERLINE:
              tags =
                g_list_append(tags, gtk_text_tag_table_lookup(tab, "underline"));
              break;
            case A_DOUBLE_UNDERLINE:
              tags =
                g_list_append(tags,
                              gtk_text_tag_table_lookup(tab, "dblunderline"));
              break;
            case A_UNDERLINE_OFF:
              tags =
                g_list_append(tags,
                              gtk_text_tag_table_lookup(tab, "underline-off"));
              break;
            case A_CROSSOUT:
              tags =
                g_list_append(tags, gtk_text_tag_table_lookup(tab, "crossout"));
              break;
            case A_CROSSOUT_OFF:
              tags =
                g_list_append(tags,
                              gtk_text_tag_table_lookup(tab, "crossout-off"));
              break;
            case A_ITALIC:
              tags =
                g_list_append(tags, gtk_text_tag_table_lookup(tab, "italic"));
              break;
            case A_ITALIC_OFF:
              tags =
                g_list_append(tags,
                              gtk_text_tag_table_lookup(tab, "italic-off"));
              break;
            case A_INVERSE:
            case A_INVERSE_OFF:
              break;              // Dunno yet
            case A_INVISIBLE:
              break;
            case A_NOWRAP:
              tags =
                g_list_append(tags, gtk_text_tag_table_lookup(tab, "nowrap"));
              break;
          }
        } else {
          // Discard the code, cause it isn't even an int
        }
      }
      // Okay newTag contains all the new codes, we don't do anything with it yet
      ptr = ansiStop + 1;
      g_strfreev(ansis);
    } else {
      // No ansi stop? Whelp! Uhm, donno, replace the escape char with an e
      *ansiStart = 'e';
    }
  }

  // Add the remaining text with the remaining TextTag
  // Text from ptr till end of the line
  world_add_text(wld, ptr, strlen(newLineTerm) - (ptr - newLineTerm), tags);
  
  world_scroll_end(wld);

  if (newLine) {
    g_free(newLine);
  }

  g_free(newLineTerm);
  g_object_get(if_main_get_widget("wndMain"), "is-active", &active, NULL);
  current = (wld == world_get_active() && gtk_notebook_get_current_page(
                    GTK_NOTEBOOK(wld->widgets.noteContainer)) == 0);

  if (current && (!GTK_WIDGET_VISIBLE(if_main_get_widget("wndMain")) || !active)) {
    wld->inactive++;
  } else if (!current) {
    wld->inactive++;

    color = g_strdup_printf("#%02x0000", CALC_COLOR_RANGE(wld->inactive));
    gdk_color_parse(color, &col);
    g_free(color);

    title =
      g_strdup_printf("%s [%d]", options_get_str(wld->settings, "name"),
                      wld->inactive);
    gtk_label_set_text(GTK_LABEL(wld->widgets.mainTabInfo.tabLabel), title);
    g_free(title);

    gtk_widget_modify_fg(wld->widgets.mainTabInfo.tabLabel, GTK_STATE_ACTIVE,
                         &col);
    gtk_image_set_from_pixbuf(GTK_IMAGE(wld->widgets.mainTabInfo.tabIcon),
                              gnoe_pixbuf_get("world_inactive.png"));

    if (wld->reColor) {
      g_source_remove(wld->reColor);
    }

    wld->reColor = g_timeout_add(1000, (GSourceFunc) on_world_recolor, wld);
  }

  tray_update();
}

/** \ingroup world
  * \brief processes a line
  *
  * \b processes a line of input for the world. It handles out of
  * bound, editing and normal text
  * @param wld the world to process the line for
  * @param line the line to process
  */
void
world_process_line(world * wld, unsigned char *line) {
  unsigned char *newLine;
  unsigned char *process;

  // First check if there is still something in the input_buffer
  // cause that should be glueed in front of the received line
  if (wld->net.input_buffer) {
    newLine = g_strconcat(wld->net.input_buffer, line, NULL);
    g_free(wld->net.input_buffer);
    wld->net.input_buffer = NULL;
  } else {
    newLine = g_strdup(line);
  }

  // Now remove the possible normal ansi, because it can happen also with mcp lines
  // Don't like it, but well
  if (strncmp(newLine, "\x1B[0m", 4) == 0) {
    // Strip it
    process = g_strdup(newLine + 4);
  } else {
    process = g_strdup(newLine);
  }

  g_free(newLine);

  if (wld->editing) {
    if (strcmp(process, ".") == 0) {
      editor_oob_end(wld);
    } else {
      editor_oob_add(wld, process);
    }
  } else if (strncmp(process, "#$#", 3) == 0) { // MCP
    if (strncasecmp(process + 3, " edit ", 6) == 0) {
      editor_invoke_oob(wld, process + 9);
    } else {
      mcp_handle_out_of_bound(wld, process + 3);  // Is this dirty? Probably :)
      world_log(wld, LOG_MCP_IN, process);
    }
  } else {
    if (strncmp(process, "#$\"", 3) == 0) {
      world_writeln(wld, process + 3);  // Is this dirty? Probably :)
      world_log(wld, LOG_IN, process + 3);
    } else {
      world_writeln(wld, process);
      world_log(wld, LOG_IN, process);
    }
  }

  g_free(process);
}

/** \ingroup world
  * \brief called whenever the world is activated
  *
  * the function is called whenever the world is activated.
  * It grabs focus and modifies the GUI to match the worlds state
  * @param wld the world which is activated
  */
void
world_activated(world * wld) {

  doEvents();

  gtk_widget_grab_focus(wld->widgets.txtVwInput);

  wld->inactive = 0;
  gtk_widget_modify_fg(wld->widgets.mainTabInfo.tabLabel, GTK_STATE_ACTIVE,
                       NULL);
  gtk_label_set_text(GTK_LABEL(wld->widgets.mainTabInfo.tabLabel),
                     options_get_str(wld->settings, "name"));
  gtk_image_set_from_pixbuf(GTK_IMAGE(wld->widgets.mainTabInfo.tabIcon),
                            gnoe_pixbuf_get("world.png"));
                            
  if_main_update_connect_button(wld->net.state == NET_CONNECTED);
}

/** \ingroup world
  * \brief \b PRIVATE: creates a color GtkTextTag
  *
  * \b PRIVATE: creates a color tag for world with the name \i colname.
  * The name is looked up in the main settings to retreive the color.
  * Usually called from \link world::world_create_tags 
  * world_create_tags \endlink
  * @param wld the world to create the tag for
  * @param colname the name of the color to create the tag for
  * @see world_create_tags
  */
void
world_create_color_tag(world * wld, gchar * colname) {
  GtkTextBuffer *buf = GTK_TEXT_BUFFER(wld->widgets.bufVwWorld);
  GdkColor col;
  options *settings = main_get_options();
  GtkTextTag *tag;
  
  gdk_color_parse(options_get_str(settings, colname), &col);

  if (colname[0] == 'f') {
    tag = gtk_text_buffer_create_tag(buf, colname, "foreground-gdk", &col, NULL);
    
    if (colname[strlen(colname) - 1] == 'h') {
      // Color tag for bold colors
      g_object_set(G_OBJECT(tag), "weight", PANGO_WEIGHT_NORMAL, NULL);
      
      if (!options_get_int(settings, "bold-colors")) {
        g_object_set(G_OBJECT(tag), "foreground-set", FALSE, NULL);
      }
    }
  } else {
    gtk_text_buffer_create_tag(buf, colname, "background-gdk", &col, NULL);
  }
}

static gboolean  
world_url_event_cb(GtkTextTag *tag,	GObject *object, GdkEvent *event,
			             GtkTextIter *iter,	GtkTextBuffer *buffer) {
	GtkTextIter start, end;
	gchar *str;

	/* If the link is being selected, don't do anything. */
	gtk_text_buffer_get_selection_bounds(buffer, &start, &end);
	
	if (gtk_text_iter_get_offset (&start) != gtk_text_iter_get_offset (&end)) {
		return FALSE;
	}

	if (event->type == GDK_BUTTON_RELEASE && event->button.button == 1) {
		start = end = *iter;
		
		if (gtk_text_iter_backward_to_tag_toggle (&start, tag) &&
		    gtk_text_iter_forward_to_tag_toggle (&end, tag)) {
			str = gtk_text_buffer_get_text (buffer,	&start,	&end, FALSE);
			gnoemoe_open_url(str);
			g_free(str);
		}
	}
	
	return FALSE;
}

static gboolean
world_view_event_cb(GtkTextView *view, GdkEventMotion *event, GtkTextTag *tag) {
	static GdkCursor *hand = NULL;
	static GdkCursor *beam = NULL;
	GtkTextWindowType type;
	GtkTextIter iter;
	GdkWindow *win;
	gint x, y, buf_x, buf_y;

	type = gtk_text_view_get_window_type(GTK_TEXT_VIEW(view), event->window);
	
	if (type != GTK_TEXT_WINDOW_TEXT) {
		return FALSE;
	}

	/* Get where the pointer really is. */
	win = gtk_text_view_get_window(GTK_TEXT_VIEW (view), type);
	gdk_window_get_pointer(win, &x, &y, NULL);
	
	/* Get the iter where the cursor is at */
	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(view), type, 
					                              x, y, &buf_x, &buf_y);
	
	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(view), &iter, buf_x, buf_y);

	if (!hand) {
		hand = gdk_cursor_new (GDK_HAND2);
		beam = gdk_cursor_new (GDK_XTERM);
	}
	
	if (gtk_text_iter_has_tag(&iter, tag)) {
		gdk_window_set_cursor (win, hand);
	} else {
		gdk_window_set_cursor (win, beam);
	}
	
	return FALSE;
}

/** \ingroup world
  * \brief \b PRIVATE: create all the GtkTextTags
  *
  * \b PRIVATE: create all color and style tags. This is only called once
  * for each world as the tags only need to be created once. Later on they only
  * need to be updated.
  * Usually called from \link world::world_apply_ansi world_apply_ansi \endlink
  * @param wld the world to create the tags for
  * @see world_apply_ansi
  */
void
world_create_tags(world * wld) {
  GtkTextBuffer *buf = GTK_TEXT_BUFFER(wld->widgets.bufVwWorld);
  GtkTextTag *tag;
  int i;

  // Create all the color tags
  for (i = 0; i < sizeof(ansi_colors) / sizeof(ansinamepair); i++) {
    world_create_color_tag(wld, (gchar *) ansi_colors[i].name);
  }

  // Bold tags
  if (options_get_int(main_get_options(), "bold-colors")) {
    tag = gtk_text_buffer_create_tag(buf, "bold", "weight", 
                                     PANGO_WEIGHT_NORMAL,
                                     NULL);
  } else {  
    tag = gtk_text_buffer_create_tag(buf, "bold", "weight", 
                                     PANGO_WEIGHT_ULTRABOLD,
                                     NULL);
  }
  
  gtk_text_buffer_create_tag(buf, "faint", "weight", PANGO_WEIGHT_ULTRALIGHT,
                             NULL);
  gtk_text_buffer_create_tag(buf, "bold-off", "weight", PANGO_WEIGHT_NORMAL,
                             NULL);

  // Underline tags
  gtk_text_buffer_create_tag(buf, "underline", "underline",
                             PANGO_UNDERLINE_SINGLE, NULL);
  gtk_text_buffer_create_tag(buf, "dblunderline", "underline",
                             PANGO_UNDERLINE_DOUBLE, NULL);
  gtk_text_buffer_create_tag(buf, "underline-off", "underline",
                             PANGO_UNDERLINE_NONE, NULL);

  // Crossout tags
  gtk_text_buffer_create_tag(buf, "crossout", "strikethrough", TRUE, NULL);
  gtk_text_buffer_create_tag(buf, "crossout-off", "strikethrough", FALSE,
                             NULL);

  // Italic tags
  gtk_text_buffer_create_tag(buf, "italic", "style", PANGO_STYLE_ITALIC, NULL);
  gtk_text_buffer_create_tag(buf, "italic-off", "style", PANGO_STYLE_NORMAL,
                             NULL);

  // Nowrap tag
  gtk_text_buffer_create_tag(buf, "nowrap", "wrap-mode", GTK_WRAP_NONE, NULL);
  
  // Url tag
  tag = gtk_text_buffer_create_tag(buf, "url", "foreground", "steelblue", 
                                   "underline", PANGO_UNDERLINE_SINGLE, NULL);
	g_signal_connect(tag, "event", G_CALLBACK (world_url_event_cb), buf);
  g_signal_connect(wld->widgets.txtVwWorld, "event", 
                   G_CALLBACK(world_view_event_cb), tag);
}

/** \ingroup world
  * \brief \b PRIVATE: update a GtkTextTag color tag
  *
  * \b PRIVATE: updates the GtkTextTag color tag matching \i name. The 
  * appropriate color is looked up in the main settings
  * Usually called from \link world::world_update_tags world_update_tags 
  * \endlink
  * @param wld the world to update the color tag for
  * @param name the name of the color tag
  * @see world_update_tags
  */
void
world_update_color_tag(world * wld, gchar * name) {
  GtkTextTagTable *tab =
    gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(wld->widgets.bufVwWorld));
  GtkTextTag *tag = gtk_text_tag_table_lookup(tab, name);
  GdkColor col;
  options *settings = main_get_options();
  
  gdk_color_parse(options_get_str(settings, name), &col);

  if (name[0] == 'f') {
    if (name[strlen(name) - 1] == 'h') {
      // Color tag for bold colors
      if (options_get_int(settings, "bold-colors")) {
        g_object_set(G_OBJECT(tag), "foreground-gdk", &col, "foreground-set",
                     TRUE, NULL);
      } else {
        g_object_set(G_OBJECT(tag), "foreground-gdk", &col, "foreground-set",
                     FALSE, NULL);
      }
    } else {
      g_object_set(G_OBJECT(tag), "foreground-gdk", &col, NULL);
    }
  } else {
    g_object_set(G_OBJECT(tag), "background-gdk", &col, NULL);
  }
}

/** \ingroup world
  * \brief \b PRIVATE: update a GtkTextTag color tag
  *
  * \b PRIVATE: updates the GtkTextTag color tag matching \i name. The 
  * appropriate color is looked up in the main settings.
  * Usually called from \link world::world_apply_ansi world_apply_ansi \endlink
  * @param wld the world to update the color tag for
  * @see world_apply_ansi
  */
void
world_update_tags(world * wld) {
  GtkTextTagTable *tab =
    gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(wld->widgets.bufVwWorld));
  GtkTextTag *tag;
  int i;

  // Update all the color tags
  for (i = 0; i < sizeof(ansi_colors) / sizeof(ansinamepair); i++) {
    world_update_color_tag(wld, (gchar *) ansi_colors[i].name);
  }
  
  tag = gtk_text_tag_table_lookup(tab, "bold");
  
  if (options_get_int(main_get_options(), "bold-colors")) {
    g_object_set(G_OBJECT(tag), "weight", PANGO_WEIGHT_NORMAL, NULL);
  } else {
    g_object_set(G_OBJECT(tag), "weight", PANGO_WEIGHT_ULTRABOLD, NULL);
  }

}

/** \ingroup world
  * \brief updates the ansi for the world
  *
  * updates all ansi tags for the world. The ansi settings are
  * looked up from the main settings and applied
  * @param wld the world to update the ansi for
  * @see world_create_tags
  * @see world_update_tags
  */
void
world_apply_ansi(world * wld) {
  GdkColor col;
  PangoFontDescription *f;
  GtkTextBuffer *buf = GTK_TEXT_BUFFER(wld->widgets.bufVwWorld);
  GtkTextTagTable *tab = gtk_text_buffer_get_tag_table(buf);
  options *settings = main_get_options();
  
  // Apply standard foreground
  gdk_color_parse(options_get_str(settings, "fg_default"), &col);
  gtk_widget_modify_text(wld->widgets.txtVwWorld, GTK_STATE_NORMAL, &col);

  // Apply standard background
  gdk_color_parse(options_get_str(settings, "bg_default"), &col);
  gtk_widget_modify_base(wld->widgets.txtVwWorld, GTK_STATE_NORMAL, &col);

  // Apply standard font
  f =
    pango_font_description_from_string(options_get_str
                                       (settings, "font-family"));
  
  if (f) {
    gtk_widget_modify_font(wld->widgets.txtVwWorld, f);
    gtk_widget_modify_font(wld->widgets.txtVwInput, f);
    pango_font_description_free(f);
  }

  world_update_character_metrics(wld);

  if (gtk_text_tag_table_get_size(tab) == 0) {
    // Create the tags
    world_create_tags(wld);
  } else {
    world_update_tags(wld);
  }
  
  // Update background
  world_update_background(wld);
}

static void
world_open_address_cb(GtkMenuItem *menuitem, const gchar *url)
{
	gnoemoe_open_url(url);
}

static void
world_copy_address_cb (GtkMenuItem *menuitem, const gchar *url)
{
	GtkClipboard *clipboard;

	clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
	gtk_clipboard_set_text(clipboard, url, -1);
}

static void
world_view_populate_popup(GtkTextView *view, GtkMenu *menu, world *wld) {
	GtkTextTagTable *table;
	GtkTextTag *tag;
	gint x, y;
	GtkTextIter iter, start, end;
	GtkWidget *item;
	gchar *str = NULL;

	table = gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(wld->widgets.bufVwWorld));
	tag = gtk_text_tag_table_lookup(table, "url");

	gtk_widget_get_pointer(GTK_WIDGET(view), &x, &y);
	
	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(view), 
					                              GTK_TEXT_WINDOW_WIDGET,
					                              x, y,
					                              &x, &y);
	
	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(view), &iter, x, y);
	start = end = iter;
	
	if (gtk_text_iter_backward_to_tag_toggle(&start, tag) &&
	    gtk_text_iter_forward_to_tag_toggle(&end, tag)) {				
		str = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(wld->widgets.bufVwWorld), 
		                               &start, &end, FALSE);
	}

  if (!str || strlen(str) == 0) {
		return;
	}

	item = gtk_menu_item_new();
	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), item);
	gtk_widget_show(item);
	
	/* Set data just to get the string freed when not needed. */
	g_object_set_data_full(G_OBJECT(menu), "url", str, (GDestroyNotify)g_free);

	item = gtk_menu_item_new_with_mnemonic(_("_Copy Link Address"));
	g_signal_connect(item, "activate", G_CALLBACK(world_copy_address_cb), str);
	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), item);
	gtk_widget_show(item);

	item = gtk_menu_item_new_with_mnemonic(_("_Open Link"));
	g_signal_connect(item, "activate", G_CALLBACK(world_open_address_cb), str);
	gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), item);
	gtk_widget_show (item);
}

/** \ingroup world
  * \brief load the world
  *
  * loads the world into the main application (adding a child to the main
  * notebook) and starts connecting
  * @param wld the world to load
  */
void
world_load(world * wld) {
  GtkWidget *noteMain = if_main_get_widget("noteMain");
  GtkWidget *child;
  GtkWidget *lbl;
  GdkCursor *cursor;
  GList *w;
  
  debug_msg(1, "world_load: loading world %s",
            options_get_str(wld->settings, "name"));

  if (!wld->loaded) {
    // Create the GUI
    child = create_world_view(wld);

    // Create the main label for this world, and connect the signal of the unload world button 
    lbl =
      create_tab_label("world.png", options_get_str(wld->settings, "name"),
                       TRUE, &(wld->widgets.mainTabInfo));
    g_signal_connect((gpointer) (wld->widgets.mainTabInfo.tabExitButton),
                     "released", G_CALLBACK(on_tabExitButton_released), wld);

    g_signal_connect(wld->widgets.txtVwWorld, "populate_popup", 
                     G_CALLBACK(world_view_populate_popup), wld);

    // Enable interface things if there where no worlds loaded previously
    for (w = worlds; w; w = w->next) {
      if (((world *) (w->data))->loaded) {
        break;
      }
    }

    wld->loaded = 1;

    gtk_notebook_append_page(GTK_NOTEBOOK(noteMain), child, lbl);
    gtk_notebook_set_current_page(GTK_NOTEBOOK(noteMain),
                                  gtk_notebook_page_num(GTK_NOTEBOOK(noteMain),
                                                        child));

    world_apply_ansi(wld);

    if (!w) {
      interface_worlds_loaded();
    }
    
    // Change the cursor to common arrow
    cursor = gdk_cursor_new(GDK_LEFT_PTR);
    gdk_window_set_cursor(gtk_text_view_get_window
                          (GTK_TEXT_VIEW(wld->widgets.txtVwWorld),
                           GTK_TEXT_WINDOW_TEXT), cursor);
    gdk_cursor_unref(cursor);

    doEvents();

    // Settings userlist width
    gtk_paned_set_position(GTK_PANED(wld->widgets.hpane),
                           wld->widgets.hpane->allocation.width -
                           options_get_int(wld->settings, "pane_position"));

    doEvents();

    net_do_connect(wld);
    world_scroll_end_force(wld);

    wld->bufMcp = gtk_text_buffer_new(if_mcpconsole_get_tag_table());
    mcpconsole_add_world(wld);
  } else {
    gtk_notebook_set_current_page(GTK_NOTEBOOK(noteMain),
                                  gtk_notebook_page_num(GTK_NOTEBOOK(noteMain),
                                                        wld->widgets.
                                                        noteContainer));
  }
}

/** \ingroup world
  * \brief \b PRIVATE: reconnect if nescessary
  *
  * \b PRIVATE: reconnects the world if nescessary, wld->tryReconnect must be
  * TRUE && wld->net.state MUST BE NET_DISCONNECTED
  * @param wld the world to try reconnecting
  * @result returns FALSE (removal of the source from the glib main loop)
  */
gboolean
world_try_reconnect(world * wld) {
  if (wld->tryReconnect != -1 && wld->net.state == NET_DISCONNECTED) {
    wld->tryReconnect = -1;
    net_do_connect(wld);
  }

  return FALSE;
}

/** \ingroup world
  * \brief called when connected
  *
  * this function is called whenever the world is connected. Its therefor
  * usually only called from the net_do_connect function
  * @param wld the world which is connected
  */
void world_connected(world *wld) {
  gchar *plr = options_get_str(wld->settings, "player_name");
  gchar *pass = options_get_str(wld->settings, "password");
  
  #ifdef HASRUBY
    script_function *f;
    if ((f = script_find("on_connect")) != NULL) {
      script_run(wld, f, NULL);
    }
  #endif

  if (strlen(plr) > 0) {
    gchar *sendLine = g_strconcat("connect ", plr, " ", pass, NULL);

    world_writeln(wld, _("# Auto-login"));
    net_send_line(wld, sendLine);
    g_free(sendLine);

    #ifdef HASRUBY
      script_function *f;
      if ((f = script_find("on_login")) != NULL) {
        script_run(wld, f, NULL);
      }
    #endif    
  }
  
  if (world_get_active() == wld) {
    if_main_update_connect_button(TRUE);
  }
}
/** \ingroup world
  * \brief called when disconnected
  *
  * this function is called whenever the world is disconnected. Its therefor
  * usually only called from the net_do_disconnect function. This will handle
  * autoreconnect if nescessary
  * @param wld the world which is disconnected
  */
void
world_disconnected(world * wld) {
  mcp_close(wld);
  
  if (world_get_active() == wld) {
    if_main_update_connect_button(FALSE); 
  }

  if (wld->loaded) {
    #ifdef HASRUBY
      script_function *f;
      if ((f = script_find("on_disconnect")) != NULL) {
        script_run(wld, f, NULL);
      }
    #endif

    if (wld->tryReconnect == -1 && !wld->manualDisconnect
        && options_get_int(wld->settings, "reconnect")) {
      wld->tryReconnect =
        g_timeout_add(3000, (GSourceFunc) world_try_reconnect, wld);
      world_writeln(wld, _("# Reconnecting in 3 seconds"));
    }
  }
}

/** \ingroup world
  * \brief unload the world
  *
  * unloads the world from the gnoemoe main window, thereby disconnecting the
  * world if nescessary. Options are saved (pane_position also), editors are
  * closed and the widgets are destroyed
  * @param wld the world to unload
  */
void
world_unload(world * wld) {
  GList *w;

  if (wld->loaded) {
    wld->loaded = 0;
    net_do_disconnect(wld);

    options_set_int(wld->settings, "pane_position",
                    wld->widgets.hpane->allocation.width -
                    gtk_paned_get_position(GTK_PANED(wld->widgets.hpane)));
    options_save(wld->settings);

    // Destroy editors
    editors_close(wld);

    gtk_widget_destroy(wld->widgets.noteContainer);
    g_object_unref(wld->bufMcp);

    for (w = worlds; w; w = w->next) {
      if (((world *) (w->data))->loaded) {
        break;
      }
    }

    // If this was the last loaded world, then we must disable some menu items
    if (!w) {
      interface_no_worlds_loaded();
    }
  }
}

/** \ingroup world
  * \brief destroy all worlds
  *
  * destroy all the worlds in \link world::worlds \endlink and free all memory
  * allocated by the worlds
  */
void
world_destroy_worlds() {
  GList *elem;
  world *wld;

  for (elem = worlds; elem; elem = elem->next) {
    wld = (world *) (elem->data);

    world_destroy(wld);
    g_free(wld);
  }

  g_list_free(worlds);
}

/** \ingroup world
  * \brief add world
  *
  * add a new world to the worlds collection. The GUI is updated (menu item
  * added to the world menu)
  * @param newWorld the world to be added
  */
void
world_add(world * newWorld) {
  worlds = g_list_append(worlds, newWorld);
  newWorld->widgets.menuItem = interface_add_world(newWorld);
  world_update(newWorld);
}

/** \ingroup world
  * \brief remove world
  *
  * remove a world from the worlds collection. The GUI is updated (menu item
  * removed) and the world is destroyed
  * @param name the name of the world to be removed
  */
void
world_remove(gchar * name) {
  world *wld;

  wld = world_get_by_name(name);

  if (wld) {
    interface_remove_world(wld);

    // Now that we got rid of that, remove the world from the worlds list
    worlds = g_list_remove(worlds, wld);
    // Now remove the whole directory (wld->path)
    main_directory_remove_all(wld->path, TRUE);
    // And destroy it totaly
    world_destroy(wld);
  }
}

/** \ingroup world
  * \brief find world by name
  *
  * find a world by matching the name of the world
  * @param name the name of the world to find
  * @result the world which matches \i name or NULL of no world is found
  */
world *
world_get_by_name(gchar * name) {
  GList *elem;
  gchar *wldName;

  for (elem = worlds; elem; elem = elem->next) {
    wldName = options_get_str(((world *) (elem->data))->settings, "name");
    if (!g_strcasecmp(name, wldName)) {
      return (world *) (elem->data);
    }
  }

  return NULL;
}

/** \ingroup world
  * \brief updates the world
  *
  * update the worlds interface by calling interface_update_world and checks/
  * creates the logs and tmp dirs if nescessary
  * @param wld the world which is updated
  */
void
world_update(world * wld) {
  interface_update_world(wld);

  // See if there is need to create a logs dir
  gchar *logsDir = g_strconcat(wld->path, "/logs", NULL);
  gchar *tmpDir = g_strconcat(wld->path, "/tmp", NULL);

  if (!g_file_test(logsDir, G_FILE_TEST_EXISTS)) {
    mkdir(logsDir, 0750);
  } else {
    chmod(logsDir, 0750);
  }

  if (!g_file_test(tmpDir, G_FILE_TEST_EXISTS)) {
    mkdir(tmpDir, 0750);
  } else {
    chmod(tmpDir, 0750);
  }

  g_free(logsDir);
  g_free(tmpDir);
}

/** \ingroup world
  * \brief load worlds from path
  *
  * load all worlds found in \i path. The worlds are directories containing
  * the worlds information
  * @param path the path to look for worlds in
  * @param autoload if TRUE the worlds which have the autoload option set will
  * be loaded
  */
void
world_load_worlds(gchar * path, gboolean autoload) {
  // appPath must be set correctly!
  GDir *dHandle = g_dir_open(path, 0, NULL);
  char *fName, *fPath;
  world *newWorld;

  if (dHandle != NULL) {
    while ((fName = (char *) g_dir_read_name(dHandle)) != NULL) {
      fPath = g_strconcat(path, "/", fName, NULL);

      if (g_file_test(fPath, G_FILE_TEST_IS_DIR)) {
        if ((newWorld = world_create(fPath)) != NULL) {
          world_add(newWorld);
          if (autoload && options_get_int(newWorld->settings, "autoload")) {
            world_load(newWorld);
          }
        }
      }

      g_free(fPath);
    }

    g_dir_close(dHandle);
  } else {
    debug_msg(1, "load_worlds: Failed to open worldspath %s", path);
  }
}

/** \ingroup world
  * \brief find world by page
  *
  * find a world by matching a child widget with the children of noteMain.
  * @param page the child widget or NULL to find the active world
  * @result the world which matches \i child or NULL of no world is found
  */
world *
world_get_by_page(GtkWidget * page) {
  GtkNotebook *noteMain = GTK_NOTEBOOK(if_main_get_widget("noteMain"));
  GList *child;
  world *wld;

  if (!page) {
    page =
      gtk_notebook_get_nth_page(noteMain,
                                gtk_notebook_get_current_page(noteMain));
  }

  if (page) {
    for (child = worlds; child; child = child->next) {
      wld = (world *) (child->data);
      if (wld->widgets.noteContainer == page) {
        return wld;
      }
    }
  }

  return NULL;
}

/** \ingroup world
  * \brief find world by index
  *
  * find a world by the index of the world in noteMain
  * @param p the index of the world
  * @result the world at index \i p or NULL of no world is found
  */
world *
world_get_by_tab_index(int p) {
  GtkWidget *page;

  if (p == -1) {
    page = NULL;
  } else {
    page =
      gtk_notebook_get_nth_page(GTK_NOTEBOOK(if_main_get_widget("noteMain")),
                                p);
  }

  return world_get_by_page(page);
}

/** \ingroup world
  * \brief get active world
  *
  * gets the world which is currently active
  * @result the world which is active or NULL if no world is active
  */
world *
world_get_active() {
  return world_get_by_page(NULL);
}

/** \ingroup world
  * \brief get worlds list
  *
  * gets the worlds list
  * @result the worlds list
  */
GList *
world_get_worlds() {
  return worlds;
}

/** \ingroup world
  * \brief \b PRIVATE: check if text view is bottom scrolled
  *
  * \b PRIVATE: check if the text view of the world is scrolled to the bottom.
  * Returns TRUE if the text view is bottom scrolled and FALSE if not.
  * @param wld the world to check bottom scrolled for
  * @result TRUE if bottom scrolled and FALSE if not
  */
gboolean
world_is_end_scrolled(world * wld) {
  GtkAdjustment *ad =
    gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW
                                        (wld->widgets.srlWndText));

  return ((ad->page_size + ad->value) >= ad->upper - wld->characterHeight);
}

/** \ingroup world
  * \brief check if scroll to end is needed
  *
  * checks if scroll to bottom is needed if text is added. It checks if it is
  * scrolled to the bottom NOW and sets wld->needScroll if it is. The main idea
  * is that you call this function, then add text and then call
  * \link world::world_scroll_end world_scroll_end \endlink to do the actual
  * scrolling. This way the text view will allways be nicely scrolled
  * @param wld the world to prepare to scroll its text view for
  */
void
world_scroll_end_prepare(world *wld) {
  wld->needScroll = world_is_end_scrolled(wld);
}

/** \ingroup world
  * \brief scroll to end if needed
  *
  * scroll to the bottom of the text view if needed. To check if there is need
  * for a nice scroll you should first call 
  * \link world::world_scroll_end_prepare world_scroll_end_prepare \endlink
  * to check if scroll will to bottom is needed when text is added (you only
  * have to call \link world::world_scroll_end_prepare 
  * world_scroll_end_prepare \endlink, then add text and finally call
  * \link world::world_scroll_end world_scroll_end \endlink
  * @param wld the world to scroll its text view for
  */
void
world_scroll_end(world * wld) {
  if (wld->needScroll) {
    world_scroll_end_force(wld);
    wld->needScroll = FALSE;
  }
}

/** \ingroup world
  * \brief scroll to end
  *
  * scroll to the bottom no matter what
  * @param wld the world to scroll its text view for
  */
void
world_scroll_end_force(world * wld) {
  gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(wld->widgets.txtVwWorld),
                               gtk_text_buffer_get_mark(GTK_TEXT_BUFFER
                                                        (wld->widgets.
                                                         bufVwWorld),
                                                        "end-of-buffer"), 0.0,
                               TRUE, 1.0, 1.0);
}

/** \ingroup world
  * \brief set the status bar message
  *
  * sets the status bar message to some new \i text
  * @param wld the world to change the status bar message for
  * @param text the new message
  */
void
world_update_status(world * wld, gchar * text) {
  gtk_statusbar_pop(GTK_STATUSBAR(wld->widgets.stbStatus), 0);
  gtk_statusbar_push(GTK_STATUSBAR(wld->widgets.stbStatus), 0, text);
}

/** \ingroup world
  * \brief log a line to the logfile 
  *
  * logs a line to the worlds logfile. Specify the log_type for format
  * purposes (and later on to determine if this type of log needs to be logged)
  * @param wld the world to log for
  * @param ltype the type of line
  * @param line the text line
  */
void
world_log(world * wld, log_type ltype, gchar * line) {
  FILE *f;
  GString *s;
  gchar *start, *log, *noAnsi;
  struct tm *timet;
  time_t timer;

  timer = time(0);
  timet = localtime(&timer);

  log =
    g_strdup_printf("%s/logs/%04d-%02d-%02d.log", wld->path,
                    timet->tm_year + 1900, timet->tm_mon + 1, timet->tm_mday);
  f = fopen(log, "a");
  g_free(log);

  if (!f) {
    return;
  }

  start =
    g_strdup_printf("[%02d:%02d:%02d] ", timet->tm_hour, timet->tm_min,
                    timet->tm_sec);

  s = g_string_new(start);
  g_free(start);

  switch (ltype) {
  case LOG_IN:
    s = g_string_append_c(s, '<');
    break;
  case LOG_OUT:
    s = g_string_append_c(s, '>');
    break;
  case LOG_MCP_IN:
    s = g_string_append(s, "[MCP] <");
    break;
  case LOG_MCP_OUT:
    s = g_string_append(s, "[MCP] >");
    break;
  case LOG_STATUS:
    s = g_string_append_c(s, '#');
    break;
  }

  s = g_string_append(s, " ");
  s = g_string_append(s, line);

  noAnsi = ansi_strip(g_strdup(s->str));

  fprintf(f, "%s\n", noAnsi);
  fclose(f);

  g_free(noAnsi);
  g_string_free(s, TRUE);
}

/** \ingroup world
  * \brief calculate the character width/height
  *
  * calculates the character width and height and stores these values in the
  * wld->characterWidth and wld->characterHeight
  * @param wld the world to calculate these values for
  */
void
world_update_character_metrics(world *wld) {
  GtkRcStyle *style = gtk_widget_get_modifier_style(wld->widgets.txtVwWorld);
  PangoContext *pc = gtk_widget_create_pango_context(wld->widgets.txtVwWorld);

  if (style->font_desc != NULL) {
    pango_context_set_font_description(pc, style->font_desc);
    PangoLayout *pl = pango_layout_new(pc);

    pango_layout_set_text(pl, "G", 1);
    pango_layout_get_pixel_size(pl, &(wld->characterWidth), &(wld->characterHeight));
    g_object_unref(pl);
  }
  
  g_object_unref(pc);
}

/** \ingroup world
  * \brief update the background
  *
  * updates the semi-transparant background, DON'T call this function yourself
  * this will be called only when it is needed (when the window moves etc)
  * @param wld the world to update the background for
  */
void
world_update_transparent_background(world *wld) {
  GdkWindow *wnd;
  int x,y, width, height, widtha, heighta;
  GdkAtom type;
  guchar *data;
  GdkRectangle paint;
  GdkPixmap *root, *bg;
  GdkPixbuf *b;
  GdkColor col;
  GdkGC *gc;
  
  if (wld->widgets.txtVwWorld == NULL) {
    return;
  }
  
  wnd = gtk_text_view_get_window(GTK_TEXT_VIEW(
                      wld->widgets.txtVwWorld), GTK_TEXT_WINDOW_TEXT);
  
  if (wnd == NULL) {
    return;
  }
  
  gdk_window_get_deskrelative_origin(wnd, &x, &y);
  gdk_window_get_geometry(wnd, NULL, NULL, &width, &height, NULL);

  gdk_property_get(GDK_ROOT_PARENT(), gdk_atom_intern("_XROOTPMAP_ID", TRUE),
		     0, 0, 10, FALSE, &type, NULL, NULL, &data);

  if (!data) {
    return;
  }
  
  root = gdk_pixmap_foreign_new(*((Pixmap *)data));
  gdk_drawable_get_size(GDK_DRAWABLE(root), &widtha, &heighta);
  
  if (x + width > widtha) {
    width = widtha - x;
  }
  
  if (y + height > heighta) {
    height = heighta - y;
  }

	b = gdk_pixbuf_get_from_drawable(NULL, GDK_DRAWABLE(root), gdk_colormap_get_system(),
	                 x, y, 0, 0, width, height);
  
  bg = gdk_pixmap_new(GDK_DRAWABLE(wnd), width, height, -1);
  gc = gdk_gc_new(GDK_DRAWABLE(wnd));
  
  gnoe_pixbuf_set_alpha(&b, options_get_int(main_get_options(), "background_transparancy"));
  gdk_color_parse(options_get_str(main_get_options(), "bg_default"), &col);
  gdk_gc_set_rgb_fg_color(gc, &col);
  
  gdk_draw_rectangle(GDK_DRAWABLE(bg), gc, TRUE, 0, 0, width, height);
  gdk_draw_pixbuf(GDK_DRAWABLE(bg), gc, b, 0, 0, 0, 0, width, height, GDK_RGB_DITHER_NONE, 0, 0);
  
  gdk_pixbuf_unref(b);
  gdk_gc_unref(gc);
  gdk_window_set_back_pixmap(wnd, bg, FALSE);
  
  paint.x = 0; paint.y = 0; paint.width = width; paint.height = height;
  
  gdk_window_invalidate_region(wnd, gdk_region_rectangle(&paint), TRUE);
  gdk_window_process_updates(wnd, TRUE);
}

/** \ingroup world
  * \brief the update background timeout callback
  *
  * callback used to minimize the times the background is updated
  * @param wld the world to update the background for
  */
gboolean
world_update_background_cb(world * wld) {
  if (wld->updateBg != -1) {
    wld->updateBg = -1;
    world_update_transparent_background(wld);
  }
  
  return FALSE;
}

/** \ingroup world
  * \brief update the background
  *
  * this function is called when the background needs to be updated. It is 
  * implemented to only update the background once in 500 milliseconds
  * @param wld the world to update the background for
  */
void world_update_background(world *wld) {
  if (options_get_int(main_get_options(), "background_transparent")) {
    if (wld->updateBg == -1) {
      g_source_remove(wld->updateBg);
    }
    
    wld->updateBg = g_timeout_add(500, (GSourceFunc) world_update_background_cb, wld);
  }
}

/** \ingroup world
  * \brief process an input line 
  *
  * processes a line of input, usually called when text is entered in the input
  * widget. This function will do the appropriate thing (such as running a script
    or just sending the line)
  * @param wld the world to process input for
  * @param text the text to be processed
  */
void
world_process_input(world *wld, gchar *text) {
  #ifdef HASRUBY
    gchar *space, *script, *argstr;
    script_function *f;
    gboolean scriptRan = FALSE;

    if (strlen(text) > 1) {
      if (text[0] == '/' && text[1] != '/') {
        space = strstr(text, " ");
        
        if (space == NULL) {
          script = g_strdup(text + 1);
          argstr = g_strdup("");
        } else {
          script = g_strndup(text + 1, (space - text) - 1);
          argstr = g_strdup(space + 1);
        }
        
        debug_msg(1, "world_process_input: find script %s", script);
        
        if ((f = script_find(script)) != NULL) {
          scriptRan = script_run(wld, f, argstr);
        }
        
        g_free(script);
        g_free(argstr);
        
        if (scriptRan) {
          return;
        }
      } else if (text[0] == '/' && text[1] == '/') {
        text = text + 1;
      }
    }
  #endif
    
  net_send_line(wld, text);
  world_log(wld, LOG_OUT, text);
}

void
world_userlist_create(world *wld) {
  GtkTreeView *userlist = GTK_TREE_VIEW(wld->widgets.tvwUserlist);
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;
  GtkListStore *store = 
    gtk_list_store_new(WORLD_USERLIST_N_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING, 
                       G_TYPE_INT, G_TYPE_STRING);
  GtkTreeModel *model =
    gtk_tree_model_sort_new_with_model(GTK_TREE_MODEL(store));
  gtk_tree_view_set_model(userlist, model);

  renderer = gtk_cell_renderer_pixbuf_new();
  column =
    gtk_tree_view_column_new_with_attributes(_("I"), renderer, "pixbuf",
                                             WORLD_USERLIST_ICON, NULL);
  gtk_tree_view_append_column(userlist, column);

  renderer = gtk_cell_renderer_text_new();
  column =
    gtk_tree_view_column_new_with_attributes(_("Name"), renderer, "text",
                                             WORLD_USERLIST_NAME, NULL);
  gtk_tree_view_append_column(userlist, column);

  renderer = gtk_cell_renderer_text_new();
  column =
    gtk_tree_view_column_new_with_attributes(_("Obj"), renderer, "text",
                                             WORLD_USERLIST_OBJ, NULL);
  gtk_tree_view_column_set_visible(column, FALSE);
  gtk_tree_view_append_column(userlist, column);

  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model), 
                                       WORLD_USERLIST_SORT, GTK_SORT_ASCENDING);
  gtk_widget_show(GTK_WIDGET(userlist));
  GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(userlist), GTK_CAN_FOCUS);
}

void
world_userlist_free(world *wld) {
  GtkTreeModel *model, *store;
  GtkTreeView *userlist = GTK_TREE_VIEW(wld->widgets.tvwUserlist);
  GtkTreeViewColumn *column;
  int col;
  
  if (userlist) {
    model = gtk_tree_view_get_model(userlist);
    store = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(model));
    
    gtk_list_store_clear(GTK_LIST_STORE(store));

    col = g_list_length(gtk_tree_view_get_columns(userlist));

    while (col) {
      column = gtk_tree_view_get_column(userlist, col - 1);
      col = gtk_tree_view_remove_column(userlist, column);
    }

    gtk_widget_hide(GTK_WIDGET(userlist));
  }
  
  if (wld->userlistScrollHandler != -1) {
    g_source_remove(wld->userlistScrollHandler);  
  }  
}

gboolean
world_userlist_find(GtkTreeModel *store, int i, GtkTreeIter *iter) {
  GtkTreeIter item;
  gboolean can;
  int o;

  can = gtk_tree_model_get_iter_first(store, &item);

  while (can) {
    gtk_tree_model_get(store, &item, WORLD_USERLIST_OBJ, &o, -1);

    if (o == i) {
      *iter = item;
      return TRUE;
    }

    can = gtk_tree_model_iter_next(store, &item);
  }

  return FALSE;
}

gboolean
world_userlist_scroll_top(world *wld) {
  GtkAdjustment *vadj =
    gtk_scrolled_window_get_vadjustment(
      GTK_SCROLLED_WINDOW(wld->widgets.srlWndUserlist));

  wld->userlistScrollHandler = -1;
  doEvents();
  gtk_adjustment_set_value(vadj, vadj->lower);

  return FALSE;
}

void
world_userlist_update_user(world *wld, world_uu *uu) {
  GtkTreeIter item;
  GtkTreeModel *model, *store;
  GtkTreeView *userlist = GTK_TREE_VIEW(wld->widgets.tvwUserlist);
  gboolean sTop;
  GtkAdjustment *vadj;
  gchar *name, *sort;
  GError *err = NULL;
  int written, read;
    
  model = gtk_tree_view_get_model(userlist);
  store = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(model));

  if (!world_userlist_find(store, uu->nr, &item)) {
    // Bluh, why do we need these stupid workarounds, always working 'round and 'round
    // Check if the scrollbar is top-scrolled
    vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(wld->widgets.srlWndUserlist));
    sTop = (vadj->value == vadj->lower);

    gtk_list_store_prepend(GTK_LIST_STORE(store), &item);

    if (sTop) {
      if (wld->userlistScrollHandler != -1) {
        gtk_timeout_remove(wld->userlistScrollHandler);
      }
      // Schedule for scroll to topping
      wld->userlistScrollHandler =
        g_timeout_add(250, (GSourceFunc) world_userlist_scroll_top, wld);
    }
  }
  
  name = g_convert_with_fallback(uu->name, strlen(uu->name), "UTF-8",
                            options_get_str(wld->settings, "charset"),
                            NULL, &read, &written, &err);

  if (!name) {
    debug_msg(1, "world_userlist_update_user: name conversion failed: %s (written %d, read %d)!",
              err->message, written, read);
    g_error_free(err);
    name = g_strdup(uu->name);
  }
  
  sort = g_convert_with_fallback(uu->sort, strlen(uu->sort), "UTF-8",
                            options_get_str(wld->settings, "charset"),
                            NULL, &read, &written, &err);

  if (!sort) {
    debug_msg(1, "world_userlist_update_user: sort conversion failed: %s (written %d, read %d)!",
              err->message, written, read);
    g_error_free(err);
    sort = g_strdup(uu->sort);
  }
  
  gtk_list_store_set(GTK_LIST_STORE(store), &item, WORLD_USERLIST_ICON, 
                     uu->icon, WORLD_USERLIST_NAME, name, 
                     WORLD_USERLIST_OBJ, uu->nr, WORLD_USERLIST_SORT, 
                     sort, -1);
  g_free(name);
  g_free(sort);
}

void
world_userlist_remove_user(world *wld, int userid) {
  GtkTreeIter iter;
  GtkTreeModel *model, *store;
  GtkTreeView *userlist = GTK_TREE_VIEW(wld->widgets.tvwUserlist);

  model = gtk_tree_view_get_model(userlist);
  store = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(model));

  if (world_userlist_find(GTK_TREE_MODEL(store), userid, &iter)) {
    gtk_list_store_remove(GTK_LIST_STORE(store), &iter);
  }
}

void
world_userlist_clear(world *wld) {
  GtkTreeModel *model, *store;
  GtkTreeView *userlist = GTK_TREE_VIEW(wld->widgets.tvwUserlist);

  model = gtk_tree_view_get_model(userlist);
  store = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(model));
  
  gtk_list_store_clear(GTK_LIST_STORE(store));
}
