#ifndef WORLD_H
#define WORLD_H 1

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "g_options.h"
#include "support.h"

/** \defgroup world */

/** \ingroup world
 *  \brief Widget container for the world
 *
 *  Contains all widgets that could be useful to be modified.
 *  Stored here because it makes it easier to look them up.
 *  
 */

typedef struct _world_widgets {
  /** \brief glade interface xml file
   *
   *  The glade interface xml file, this can be used in conjunction with
   *  glade_xml_get_widget to lookup widgets that aren't in this struct.
   */
  GladeXML *xml;

  /** \brief GtkNotebook widget
    *
    * GtkNotebook widget for editor tabs 
    */
  GtkWidget *noteContainer; 
  /** \brief GtkHPaned widget
    *
    * GtkHPaned widget which contains \link _world_widgets::srlWndText 
    * srlWndText \endlink and 
    * \link _world_widgets::slrWndUserlist slrWndUserlist \endlink .
    */
  GtkWidget *hpane; 
  
  /** \brief GtkScrolledWindow widget
    *
    * GtkScrolledWindow widget which contains \link 
    * _world_widgets::txtVwWorld txtVwWorld \endlink .
    */
  GtkWidget *srlWndText; 

  /** \brief GtkTextView widget
    *
    * GtkTextView widget which is the main output text widget for the world.
    */  
  GtkWidget *txtVwWorld;
  
  /** \brief GtkTextBuffer object
    *
    * GtkTextBuffer object. It contains the buffer of \link 
    * _world_widgets::txtVwWorld txtVwWorld \endlink .
    */
  GObject *bufVwWorld;

  /** \brief GtkTreeView widget
    *
    * GtkTreeView widget which is used to display 
    * the player listing of the world.
    */
  GtkWidget *tvwUserlist;
  
  /** \brief GtkScrolledWindow widget
    *
    * GtkScrolledWindow widget which contains \link 
    * _world_widgets::tvwUserlist tvwUserlist \endlink .
    */
  GtkWidget *srlWndUserlist;
  
  /** \brief GtkTextView widget
    *
    * GtkTextView widget which is used for the input 
    */
  GtkWidget *txtVwInput;
    
  /** \brief GtkStatusbar widget
    *
    * GtkStatusbar widget which is used to display world specific messages
    */
  GtkWidget *stbStatus;

  /** \brief GtkToolbar widget
    *
    * GtkToolbar widget for future purposes
    */
  GtkWidget *tlbStatusButtons;

  /** \brief struct which contains world tab widgets
    *
    * struct which contains world tab widgets such as the icon, label and 
    *  exit button.
    */
  label_info mainTabInfo;

  /** \brief GtkMenuItem widget
    *
    * GtkMenuItem widget, reference to the worlds menuitem (in the 
    * world menu)
    */
  GtkWidget *menuItem;
} world_widgets;

/** \ingroup world
 *  \brief enum for world connection state
 *
 *  Enumeration which indicates the worlds connection state
 */
typedef enum _net_state { 
  NET_CONNECTED, /**< socket connected */
  NET_DISCONNECTED, /**< socket disconnected */
  NET_CONNECTING, /**< socket still connecting */
  NET_DISCONNECTING /**< socket still disconnecting */
} net_state;

/** \ingroup world
 *  \brief struct for world network properties
 *
 *  Struct which contains various fields for storing network properties
 */
typedef struct _net_properties {
  int socket; /**< the connection socket, is -1 if not connected */
  GIOChannel *channel; /**< the channel which is used to 'listen' on the socket via the glib main loop */
  guint source; /**< the id of the socket watch */

  unsigned char *input_buffer; /**< pending input */
  int tn_last; /**< used for telnet */
  int tn_subneg; /**< used for telnet */

  struct addrinfo *addr;
  struct addrinfo *current;

  net_state state; /**< state of the connection */
} net_properties;

/** \ingroup world
 *  \brief struct for world input information
 *
 *  Struct which contains various fields for storing input information such as
 *  the history, previous pressed key, current position in the history and
 *  the current prefix.
 */
typedef struct _entry_info {
  GList *buffer; /**< list of gchar * which represents the input history */
  guint prevKey; /**< the previous pressed key */
  guint curPos; /**< the current position in the history */
  
  /** the current search prefix, this is used to autocomplete from the history
  */
  gchar *prefix; 
} entry_info;

/** \ingroup world
 *  \brief struct for world editing information
 *
 *  Struct which contains various fields for storing editing information. This
 *  is only used if editing via mcp_simpleedit can't be used.
 */
typedef struct _editing_info {
  GList *lines; /**< the lines to be edited */
  gchar *name; /**< the name of the editor */
  gchar *upload; /**< the command to send when sending the editor contents */
} editing_info;

/** \ingroup world
 *  \brief struct for the world
 *
 *  Struct which contains various fields for storing world information.
 */
typedef struct _world {
  char *path; /**< path to the world directory */
  unsigned int loaded; /**< indicates if the world is currently loaded */
  unsigned int inactive; /**< indicates how many lines aren't read yet */
  gchar *prevLine; /**< contains the previous sended line (no mcp) */

  guint tryReconnect; /**< contains the id of the reconnect timeout */
  guint updateBg; /**< contains the id of the update background timeout */
  
  /** indicates if the disconnection is manual. if this is set gnoemoe will not
      try to auto-reconnect
    */
  gboolean manualDisconnect; 

  guint reColor; /**< contains the id of the tab coloring timeout */

  options *settings; /**< contains the world settings */

  world_widgets widgets; /**< contains the widgets information for this world */
  net_properties net; /**< contains the network information for this world */

  entry_info entryBuffer; /**< contains the input information for this world */
  editing_info *editing; /**< contains the editing information for this world */
  
  unsigned int characterWidth; /**< width of one character in the text view */
  unsigned int characterHeight; /**< height of one character in the text view */
  
  GtkTextBuffer *bufMcp; /**< buffer for mcp messages */
  /** indicates if world_scroll_end should actually scroll to the bottom of 
    * the textview. world_scroll_end_prepare sets to the appropriate value 
    */
  gboolean needScroll;
  
  guint userlistScrollHandler;
} world;

/** \ingroup world
 *  \brief enum indicating log type
 *
 *  Enumeration which indicates the log type
 */
typedef enum _log_type {
  LOG_IN, /**< incoming lines */
  LOG_OUT, /**< outgoing lines */
  LOG_MCP_IN, /**< mcp incoming lines */
  LOG_MCP_OUT, /**< mcp outgoing lines */
  LOG_STATUS /**< status lines (such as connecting information) */
} log_type;

typedef struct _world_uu {
  int nr;
  gchar *name;
  gchar *sort;
  GdkPixbuf *icon;  
} world_uu;

enum {
  WORLD_USERLIST_ICON,
  WORLD_USERLIST_NAME,
  WORLD_USERLIST_OBJ,
  WORLD_USERLIST_SORT,
  WORLD_USERLIST_N_COLUMNS
};

world *world_create(char *path);
void world_destroy(world * wld);
void world_process_line(world * wld, unsigned char *line);
void world_writeln(world * wld, unsigned char *line);
world *world_copy(world * source);

void world_activated(world * wld);

void world_update(world * wld);
void world_load_worlds(gchar * path, gboolean autoload);

void world_load(world * wld);
void world_unload(world * wld);

void world_add(world * newWorld);
void world_remove(gchar * name);

world *world_get_by_name(gchar * name);
world *world_get_by_page(GtkWidget * page);
world *world_get_by_tab_index(int p);
world *world_get_active();

GList *world_get_worlds();

void world_destroy_worlds();
void world_apply_ansi(world * wld);

void world_scroll_end_force(world * wld);
void world_scroll_end_prepare(world * wld);
void world_scroll_end(world * wld);

void world_connected(world * wld);
void world_disconnected(world * wld);
void world_update_status(world * wld, gchar * text);

void world_paste(world * wld, gchar * text);
gboolean world_is_end_scrolled(world * wld);
void world_log(world * wld, log_type ltype, gchar * line);

void world_update_character_metrics(world *wld);
void world_update_background(world *wld);

void world_process_input(world *wld, gchar *input);

/* World userlist API */
void world_userlist_create(world *wld);
void world_userlist_free(world *wld);
gboolean world_userlist_find(GtkTreeModel *store, int i, GtkTreeIter *iter);
gboolean world_userlist_scroll_top(world *wld);
void world_userlist_update_user(world *wld, world_uu *uu);
void world_userlist_remove_user(world *wld, int userid);
void world_userlist_clear(world *wld);

#endif
